"""!@package ep_bolfi.EP_BOLFI.EP_BOLFI
This file contains functions to perform Expectation Propagation on
simulator models using BOLFI (Bayesian Optimization).
"""

from copy import deepcopy
import elfi
import json
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import chi2


class NDArrayEncoder(json.JSONEncoder):
    def default(self, item):
        if isinstance(item, np.ndarray):
            return item.tolist()
        return json.JSONEncoder.default(self, item)


def combine_parameters_to_try(parameters, parameters_to_try_dict):
    """!@brief Give every combination as full parameter sets.
    @param parameters
        The base full parameter set as a dictionary.
    @param parameters_to_try_dict
        The keys of this dictionary correspond to the "parameters"' keys
        where different values are to be inserted. These are given by
        the tuples which are the values of this dictionary.
    @return
        A 2-tuple where the first item is the list of all parameter set
        combinations and the second the list of the combinations only.
    """

    # Fetch the parameters to try.
    parameters_to_try = []
    for key, value in parameters_to_try_dict.items():
        parameters_to_try.append((key, value))

    def recursive_combination(last_index):
        """! Recursively get every combination of parameters to try. """
        return_list = []
        if last_index == 0:
            key = parameters_to_try[0][0]
            for value in parameters_to_try[0][1]:
                return_list.append({key: value})
        elif last_index > 0:
            key = parameters_to_try[last_index][0]
            list_of_combinations = recursive_combination(last_index-1)
            for value in parameters_to_try[last_index][1]:
                for combination in list_of_combinations:
                    return_dict = {key: value}
                    return_dict.update(combination)
                    return_list.append(return_dict)
        return return_list

    combinations = recursive_combination(len(parameters_to_try)-1)
    parameters_list = []
    for combination in combinations:
        parameters_list.append(fix_parameters(parameters)(combination))
    return (parameters_list, combinations)


def fix_parameters(parameters_to_be_fixed):
    """!@brief Returns a function which sets some parameters in advance.
    @param parameters_to_be_fixed
        These parameters will at least be a
        part of the dictionary that the returned function returns.
    @return
        The function which adds additional parameters to a
      dictionary or replaces existing parameters with the new ones.
    """

    def return_all_parameters(free_parameters):
        """!@brief Adds the 'free_parameters' to a pool of parameters.
        @param free_parameters
            A dictionary which gets added to the fixed pool of
            parameters (see the function 'fix_parameters').
        @return
            A dictionary containing free_parameters and some other
            key-value-pairs as defined by 'fix_parameters'. If a key is
            present in both, the value from 'free_parameters' is used.
        """

        return_dict = deepcopy(parameters_to_be_fixed)
        return_dict.update(free_parameters)
        return return_dict

    return return_all_parameters


class Preprocessed_Simulator:
    """!@brief Normalizes sampling to a standard normal distribution.
    In order to help BOLFI to work efficiently with the least amount of
    setup required, this class mediates between the model parameters
    and a standard normal distribution for sampling. In a sense, the
    simulator output gets transformed into covariance eigenvectors.
    """

    def __init__(self,
                 simulator,
                 fixed_parameters,
                 free_parameters_names,
                 r,
                 Q,
                 experimental_data,
                 feature_extractor,
                 transform_parameters={},
                 fixed_parameter_order=None):
        """!
        @param simulator
            The function that returns results given parameters.
        @param fixed_parameters
            Dictionary of parameters that stay fixed and their values.
        @param free_parameters_names
            List of the names of parameters which shall be inferred.
        @param r
            'Q' times the mean of the distribution of free parameters.
        @param Q
            Inverse covariance matrix of free parameters, i.e. precision.
            It is used to transform the free parameters given to the
            'simulator' into the ones used in the model. Most notably, these
            univariate standard normal distributions get transformed into
            a multivariate normal distribution corresponding to Q and r.
        @param experimental_data
            The experimental data that the model will be fitted to.
            It has to have the same structure as the 'simulator' output.
        @param feature_extractor
            A function that takes the output of 'simulator' / the
            'experimental_data' and returns a list of numbers, the features.
        @param transform_parameters
            Optional transformations between the parameter space that is
            used for searching for optimal parameters and the battery
            model parameters. 'Q' and 'r' define a normal distribution in
            that search space. The keys are the names of the free
            parameters. The values are 2-tuples. The first entry is a
            function taking the search space parameter and returning the
            model parameter. The second entry is the inverse function.
        @param fixed_parameter_order
            Optional fixed parameter order. Prevents erroneous behaviour
            when the parameter dictionaries get reshuffled.
            Highly recommended.
        """

        self.simulator = simulator
        self.fixed_parameters = fixed_parameters
        self.free_parameters_names = free_parameters_names
        if fixed_parameter_order:
            self.fixed_parameter_order = fixed_parameter_order
        else:
            self.fixed_parameter_order = list(
                enumerate(self.free_parameters_names)
            )
        self.r = r
        self.Q = Q
        self.experimental_data = experimental_data
        self.feature_extractor = feature_extractor
        self.transform_parameters = transform_parameters

        ## Stores all parameter combinations that have been tried.
        self.log_of_tried_parameters = {
            name: [] for name in self.free_parameters_names
        }

        ## Extract the features from the experimental data.
        self.experimental_features = feature_extractor(experimental_data)

        ## Input dimension of the estimation task.
        self.input_dim = len(self.free_parameters_names)
        ## Output dimension of the estimation task (number of features).
        self.output_dim = len(self.experimental_features)

        ## Create a function to combine the free and fixed parameters.
        self.add_parameters = fix_parameters(self.fixed_parameters)

        ## Compute the linear transformation of parameters for which the
        # covariance of the underlying multivariate normal distribution
        # is a diagonal matrix. That is, compute the eigenvectors of Q.
        # It is more stable since Q has growing eigenvectors in
        # convergence.
        self.inv_variances, self.back_transform_matrix = np.linalg.eigh(self.Q)
        ## Variances of the model parameters.
        self.variances = self.inv_variances**(-1)
        ## Inverse of back_transform_matrix.
        self.transform_matrix = self.back_transform_matrix.T

        ## transform_matrix @ Q @ back_transform_matrix is diagonal. The
        # correct transformation for vectors v is then transform_matrix @
        # v. The product below corresponds to Q⁻¹ @ r. It is just
        # expressed in the eigenvector space of Q for efficiency.
        self.transformed_means = np.linalg.multi_dot(
            [np.diag(self.variances), self.transform_matrix, self.r]
        )

        ## Now that the multivariate normal distribution is decomposed into
        # various univariate ones, norm them to have equal variance 1.
        self.norm_factor = np.diag(np.sqrt(self.inv_variances))
        ## Inverse of norm_factor.
        self.un_norm_factor = np.diag(np.sqrt(self.variances))
        ## Expectation values of the normed univariate normal distributions.
        self.normed_means = np.matmul(self.norm_factor, self.transformed_means)

    def search_to_transformed_trial(self, search_space_parameters):
        """!@brief Transforms search space parameters to model ones.
        @param search_space_parameters
            A list of lists which each contain a single search space
            parameter sample as it is returned by the sample functions
            of ELFI. In the case of only sample, a list also works.
        @return
            A dictionary with its keys as the names of the parameters.
            Their order in the 'search_space_parameters' is given by the
            order of 'self.free_parameters_names'. The values yield the
            model parameters when passed through the functions in
            'self.transform_parameters'.
        """

        # Index of the axis along which a single sample lies.
        sample_axis = 1

        # Get the normed parameter sample.
        normed_parameters = np.apply_along_axis(
            lambda slice: slice + self.normed_means,
            sample_axis,
            search_space_parameters
        )

        # Un-norm it into the eigenvector space of Q represenation.
        transformed_parameters = np.apply_along_axis(
            lambda slice: self.un_norm_factor @ slice,
            sample_axis,
            normed_parameters)

        # Transform it backwards to the transformed model space.
        transformed_trial_parameters = np.apply_along_axis(
            lambda slice: self.back_transform_matrix @ slice,
            sample_axis,
            transformed_parameters
        )

        # Convert this into a dictionary. The ordering is still given by
        # 'self.free_parameters_names'.
        return {name: transformed_trial_parameters.T[i]
                for i, name in self.fixed_parameter_order}

    def transformed_trial_to_search(self, model_space_parameters):
        """!@brief Transforms model space parameters to model ones.
        @param model_space_parameters
            A dictionary. The keys are the 'self.free_parameters_names'
            and the values are the model parameters after applying the
            transformations given in 'self.transform_parameters'.
        @return
            A list (of lists) which each contain a single search space
            parameter sample as it is returned by the sample functions
            of ELFI. If the 'model_space_parameters' dictionary values
            are numbers, the returned value is a list. If they are
            lists, the returned value is a list of corresponding lists.
            In that case, each and every list must have the same length.
        """

        # Index of the axis along which a single sample lies.
        sample_axis = 1

        length = len(model_space_parameters[self.free_parameters_names[0]])
        transformed_trial_parameters = [
            [model_space_parameters[name][i]
                for name in self.free_parameters_names]
            for i in range(length)
        ]

        # Transform it forwards to the eigenvector space of Q.
        transformed_parameters = np.apply_along_axis(
            lambda slice: self.transform_matrix @ slice,
            sample_axis,
            transformed_trial_parameters
        )

        # Get the normed parameter sample.
        normed_parameters = np.apply_along_axis(
            lambda slice: self.norm_factor @ slice,
            sample_axis,
            transformed_parameters
        )

        # Get the search space parameters.
        search_parameters = np.apply_along_axis(
            lambda slice: slice - self.normed_means,
            sample_axis,
            normed_parameters)

        return search_parameters

    def undo_transformation(self, transformed_trial_parameters):
        """!@brief Undo the transforms in 'self.transform_parameters'.
        @param transformed_trial_parameters
            A dictionary. The keys are the 'free_parameters_names' and
            the values are the model parameters after they have been
            transformed as specified in 'self.transform_parameters'.
        @return
            The given dictionary with the values transformed back to the
            actual model parameter values.
        """

        trial_parameters = {
            name: transform[0](transformed_trial_parameters[name])
            for name, transform in self.transform_parameters.items()
        }

        return trial_parameters

    def apply_transformation(self, trial_parameters):
        """!@brief Apply the transforms in 'self.transform_parameters'.
        @param trial_parameters
            A dictionary. The keys are the 'free_parameters_names'
            and the values are the actual model parameters.
        @return
            The given dictionary with the vales transformed to the
            modified parameter space as specified in
            'self.transform_parameters'.
        """

        try:
            transformed_trial_parameters = {
                name: transform[1](trial_parameters[name])
                for name, transform in self.transform_parameters.items()
            }
        except KeyError as e:
            raise KeyError(
                "A transformation was given for a parameter that "
                + "isn't to be estimated. Original error message: " + repr(e)
            )

        return transformed_trial_parameters

    def elfi_simulator(self, *args, **kwargs):
        """!@brief A battery model simulator that can be used with ELFI.
        @param *args
            The parameters as given by the prior nodes. Their
            order has to correspond to that of the parameter
            'free_parameters' given to 'return_simulator'.
        @param **kwargs
            Keyword parameters batch_size and random_state,
            but both are unused (they just get passed by BOLFI).
        @return
            Simulated features for the given free parameters.
        """

        if kwargs["batch_size"] > 1:
            print("EP_BOLFI: batch_size > 1 is not implemented.")
            raise NotImplementedError

        trial_parameters = self.add_parameters(self.undo_transformation(
            self.search_to_transformed_trial(np.array(args))
        ))
        simulated_data = self.simulator(trial_parameters)
        simulated_features = self.feature_extractor(simulated_data)

        for name in self.free_parameters_names:
            # This line logs the tried parameters.
            self.log_of_tried_parameters[name].append(trial_parameters[name])

        return simulated_features


class Optimizer_State:
    """!@brief Handles the heuristics for the EP-BOLFI operation modes. """

    def __init__(
        self,
        input_dim,
        mcmc_chains,
        total_evidence,
        posterior_samples,
        gelman_rubin_threshold,
        ess_ratio_resample=5.0,
        ess_ratio_sampling_from_zero=-1.0,
        ess_ratio_abort=20.0,
        posterior_sampling_increase=1.2,
        model_resampling_increase=1.2,
    ):
        self.input_dim = input_dim
        self.mcmc_chains = mcmc_chains
        self.total_evidence = total_evidence
        self.posterior_samples = posterior_samples
        self.gelman_rubin_threshold = gelman_rubin_threshold
        self.ess_ratio_resample = ess_ratio_resample
        self.ess_ratio_sampling_from_zero = ess_ratio_sampling_from_zero
        self.ess_ratio_abort = ess_ratio_abort
        self.posterior_sampling_increase = posterior_sampling_increase
        self.model_resampling_increase = model_resampling_increase

        self.finished = False
        self.sampling_from_zero = False
        self.resampling = False
        self.initials = None

        self.order_of_actions = []
        for ess_ratio, action in sorted([
            (0.0, "sample"),
            (self.ess_ratio_resample, "resample"),
            (self.ess_ratio_sampling_from_zero, "sample_from_zero"),
            (self.ess_ratio_abort, "abort"),
        ]):
            if ess_ratio >= 0.0:
                self.order_of_actions.append((ess_ratio, action))
        self.current_action = "sample"
        self.verbose_actions = {
            "sample":
                "Sampling the pseudo-posterior...",
            "resample":
                "Resampling the model...",
            "sample_from_zero":
                "Sampling the pseudo-posterior from  0...",
            "abort":
                "Posterior too badly conditioned. Skipping feature.",
        }

    def calculate_next_step(self, ess_ratio, action=None):
        if action is None:
            for ess_criterion, action in self.order_of_actions:
                # By not breaking when the criterion is fulfilled and going
                # through them in ascending order, their ess ratios work as
                # checkpoints; the largest one smaller than the current ess
                # gets triggered. They are exclusive to each other.
                if ess_criterion < ess_ratio:
                    self.current_action = action
        else:
            self.current_action = action
        if self.current_action == "sample":
            self.initials = None
            self.posterior_samples *= self.posterior_sampling_increase
        elif self.current_action == "resample":
            self.initials = None
            self.total_evidence *= self.model_resampling_increase
        elif self.current_action == "sample_from_zero":
            self.initials = np.zeros([self.mcmc_chains, self.input_dim])
            self.posterior_samples *= self.posterior_sampling_increase
        elif self.current_action == "abort":
            self.initials = None


class EP_BOLFI:
    """!@brief Expectation Propagation and Bayesian Optimization.
    Sets up and runs these two algorithms to infer model parameters.
    Use the variables "Q", "r", "Q_features" and "r_features" to copy
    the state of another estimator. Do not use them in any other case.
    Always use either all of them or none of them.
    """

    def __init__(
        self,
        simulators,
        experimental_datasets,
        feature_extractors,
        fixed_parameters,
        free_parameters=None,
        initial_covariance=None,
        free_parameters_boundaries=None,
        boundaries_in_deviations=0,
        Q=None,
        r=None,
        Q_features=None,
        r_features=None,
        transform_parameters={},
        weights=None,
        display_current_feature=None,
        fixed_parameter_order=None,
    ):
        """!
        @param simulators
            A list of functions that take one argument: a dictionary of
            combined 'fixed_parameters' and 'free_parameters'. They return
            the simulated counterpart to the experimental data. Most of
            the time, one function will be sufficient. Additional
            functions may be used to combine simulators which each give a
            subset of the total experimental method.
        @param experimental_datasets
            A list of the experimental data. Each entry corresponds to
            the simulator in 'simulators' with the same index and has the
            same structure as its output.
        @param feature_extractors
            A list of functions which each take the corresponding data
            entry and return a list of numbers, which represent its
            reduced features.
        @param fixed_parameters
            Dictionary of parameters that stay fixed and their values.
        @param free_parameters
            Dictionary of parameters which shall be inferred and their
            initial guesses or, more accurately, their expected values.
            Please note that these values live in the transformed space.
            Optionally, the values may be a 2-tuple where the second entry
            would be the variance of that parameter. For finer tuning with
            covariances, use 'initial_covariance' (will take precedence).
            Alternatively, you may set 'free_parameters_boundaries' to
            set the expected values and variances by confidence intervals.
        @param initial_covariance
            Initial covariance of the parameters. Has to be a symmetric
            matrix (list of list or numpy 2D array). A reasonable simple
            choice is to have a diagonal matrix and set the standard
            deviation σᵢ of each parameter to half of the distance between
            initial guess and biggest/smallest value that shall be tried.
            If the diagonal entries are σᵢ² and the bounds are symmetric,
            the probability distribution of each parameter is 95%
            within these bounds. Please note that the same does not hold
            for the whole probability distribution.
        @param free_parameters_boundaries
            Optional hard boundaries of the space in which optimal
            parameters are searched for. They are given as a dictionary
            with values as 2-tuples with the left and right boundaries.
            Boundaries need to be given for either none or all parameters.
            If None are given, boundaries will be set by
            'boundaries_in_deviations' relative to the covariance.
            The default then is the 95 % confidence ellipsoid.
            If neither 'initial_covariance' nor 'free_parameters' set the
            covariance, this parameter sets it according to the example in
            the description of 'initial_covariance'.
        @param boundaries_in_deviations
            When <= 0, the boundaries are set as described above.
            When > 0, the boundaries are this multiple of the standard
            deviation. This scales with the shrinking covariance
            as the algorithm progresses. 'free_parameters_boundaries'
            takes precedence, i.e., the covariance gets set and then the
            boundaries in the optimization are in standard deviations.
        @param Q
            If you want to restore a previous EP-BOLFI instance from a
            dump of its data (see "result_to_json" method), put the
            "Q" attribute stored therein into this parameter.
        @param r
            Same as Q, but use the "r" attribute.
        @param Q_features
            Same as Q, but use the "Q_features" attribute.
        @param r_features
            Same as Q, but use the "r_features" attribute.
        @param transform_parameters
            Optional transformations between the parameter space that is
            used for searching for optimal parameters and the model
            parameters. Any missing free parameter is not transformed.
            The values are 2-tuples. The first entry is a function taking
            the search space parameter and returning the model parameter.
            The second entry is the inverse function.
            For convenience, any value may also be one of the following:
            - 'none' => (identity, identity)
            - 'log' => (exp, log)
            Please note that, for performance reasons, the returned inferred
            values are directly back-transformed from the mean of the
            internal standard distribution. This means that they represent
            the median of the actual distribution.
        @param weights
            Optional weights to assign to the features. Higher weights
            give more importance to a feature and vice versa. A list of
            lists which correspond to the 'feature_extractors'.
        @param display_current_feature
            A list of functions. Each corresponds to a feature extractor
            with the same index. Given an index of the array of its
            features, this returns a short description of it.
            If None is given, only the index will be shown in the output.
        @param fixed_parameter_order
            Establish a numerical order to the parameter names. This prevents
            errors arising from internal reordering of the dictionaries.
            Only necessary when using the same model in different contexts.
        """

        self.simulators = simulators
        self.experimental_datasets = experimental_datasets
        self.feature_extractors = feature_extractors
        self.fixed_parameters = fixed_parameters
        self.free_parameters = free_parameters
        self.initial_covariance = initial_covariance
        self.free_parameters_boundaries = free_parameters_boundaries
        self.boundaries_in_deviations = boundaries_in_deviations
        self.transform_parameters = transform_parameters
        self.weights = weights
        self.display_current_feature = display_current_feature

        if self.free_parameters:
            self.fixed_parameter_order = list(
                enumerate(self.free_parameters.keys())
            )
        elif self.free_parameters_boundaries:
            self.fixed_parameter_order = list(
                enumerate(self.free_parameters_boundaries.keys())
            )
        else:
            raise ValueError(
                "Either 'free_parameters' or 'free_parameters_boundaries' "
                "needs to be set!"
            )
        ## If the parameter order is explicitly given, use that instead.
        if fixed_parameter_order:
            self.fixed_parameter_order = fixed_parameter_order
        ## Stores all parameter combinations that have been tried.
        self.log_of_tried_parameters = {
            name: [] for _, name in self.fixed_parameter_order
        }

        ## Experimental features.
        self.experimental_features = []
        ## Input dimension of the estimation task.
        self.input_dim = len(self.log_of_tried_parameters)
        ## Output dimension of the estimation task (sum of features).
        self.output_dim = 0
        ## Mapping of index by all features to corresponding simulator.
        self.simulator_index_by_feature = []
        ## Mapping of index by all features to that by one set of them.
        self.sub_index_by_feature = []
        for i, (extractor, dataset) in (enumerate(zip(
                self.feature_extractors, self.experimental_datasets))):
            self.experimental_features.append(extractor(dataset))
            length = len(self.experimental_features[-1])
            for j in range(self.output_dim, self.output_dim + length):
                self.simulator_index_by_feature.append(i)
                self.sub_index_by_feature.append(j - self.output_dim)
            self.output_dim = self.output_dim + length
        ## Set the weights to unity if None are given.
        if self.weights is None:
            self.weights = []
            for features in self.experimental_features:
                self.weights.append([1 for _ in range(len(features))])
        ## Container for the initial expectation values.
        self.initial_guesses = np.zeros(self.input_dim)

        if self.display_current_feature is not None:
            ## Stores all raw parameter evaluation points.
            self.log_of_raw_tried_parameters = {
                self.display_current_feature[
                    self.simulator_index_by_feature[j]
                ](self.sub_index_by_feature[j]): []
                for j in range(self.output_dim)
            }
            ## Stores all discrepancies of the sampled parameters.
            self.log_of_discrepancies = {
                self.display_current_feature[
                    self.simulator_index_by_feature[j]
                ](self.sub_index_by_feature[j]): []
                for j in range(self.output_dim)
            }
        else:
            self.log_of_raw_tried_parameters = {
                str(j): [] for j in range(self.output_dim)
            }
            self.log_of_discrepancies = {
                str(j): [] for j in range(self.output_dim)
            }

        if (boundaries_in_deviations <= 0):
            self.boundaries_in_deviations = np.sqrt(
                chi2(self.input_dim).ppf(0.95)
            )
        else:
            ## Uses boundaries as multiples of the standard deviation.
            self.boundaries_in_deviations = boundaries_in_deviations

        # Substitute transformations given by name.
        if self.transform_parameters is not {}:
            for name, function in self.transform_parameters.items():
                if type(function) is str:
                    if function == 'none':
                        self.transform_parameters[name] = (
                            lambda s: s, lambda b: b
                        )
                    elif function == 'log':
                        self.transform_parameters[name] = (
                            lambda s: np.exp(s), lambda b: np.log(b)
                        )
        # Fill-in 'none' transformations.
        for _, name in self.fixed_parameter_order:
            if name not in self.transform_parameters.keys():
                self.transform_parameters[name] = (
                    lambda s: s, lambda b: b
                )

        # Set up the covariance matrix if it is not given already.
        variance_info = False
        if self.initial_covariance is None and Q is None:
            # Check if the variance information is present.
            variance_info = True
            try:
                if self.free_parameters:
                    for value in self.free_parameters.values():
                        if len(value) < 2:
                            variance_info = False
                            break
                else:
                    variance_info = False
            except TypeError:
                variance_info = False
            if not variance_info and self.free_parameters_boundaries is None:
                raise ValueError(
                    "'free_parameters' has to contain 2-tuples or "
                    "'free_parameters_boundaries' has to be set if "
                    "'initial_covariance' is not given!"
                )

            self.initial_covariance = np.zeros([
                self.input_dim, self.input_dim
            ])
            if variance_info:
                for i, name in self.fixed_parameter_order:
                    value = self.free_parameters[name]
                    self.initial_guesses[i] = value[0]
                    self.initial_covariance[i][i] = value[1]
            else:
                if self.free_parameters:
                    for i, name in self.fixed_parameter_order:
                        value = self.free_parameters[name]
                        self.initial_guesses[i] = np.atleast_1d(value)[0]
                else:
                    for i, name in self.fixed_parameter_order:
                        value = self.free_parameters_boundaries[name]
                        transform = self.transform_parameters[name]
                        expected_value = 0.5 * (
                            transform[1](value[0]) + transform[1](value[1])
                        )
                        self.initial_guesses[i] = expected_value
                for i, name in self.fixed_parameter_order:
                    value = self.free_parameters_boundaries[name]
                    transform = self.transform_parameters[name]
                    standard_deviation = np.max([
                        np.abs(transform[1](value[1])
                               - self.initial_guesses[i]),
                        np.abs(self.initial_guesses[i]
                               - transform[1](value[0]))
                    ]) / self.boundaries_in_deviations
                    self.initial_covariance[i][i] = standard_deviation**2
        else:
            if self.free_parameters:
                for i, name in self.fixed_parameter_order:
                    value = self.free_parameters[name]
                    self.initial_guesses[i] = np.atleast_1d(value)[0]
            else:
                raise ValueError(
                    "When 'initial_covariance' or 'Q' are set, "
                    "'free_parameters' needs to be set as well!"
                )
        if Q is not None:
            self.initial_covariance = np.linalg.inv(Q)

        ## Stores the inference mean (empty at first).
        self.final_expectation = deepcopy(self.initial_guesses)
        ## Stores the inference covariance (empty at first).
        self.final_covariance = deepcopy(self.initial_covariance)
        deviations = np.sqrt(np.diag(self.final_covariance))
        ## Stores the inference correlation (empty at first).
        self.final_correlation = (
            self.final_covariance / deviations[:, None]
        ) / deviations[None, :]

        if Q is None:
            ## Expectation Propagation covariance matrix (prior).
            self.initial_Q = np.linalg.inv(self.initial_covariance)
            ## Expectation Propagation covariance matrix (posterior).
            self.Q = deepcopy(self.initial_Q)
        else:
            ## Expectation Propagation covariance matrix (prior).
            self.initial_Q = deepcopy(Q)
            ## Expectation Propagation covariance matrix (posterior).
            self.Q = deepcopy(Q)
        if r is None:
            ## Expectation Propagation expectation value (prior).
            self.initial_r = self.initial_Q @ self.initial_guesses
            ## Expectation Propagation expectation value (posterior).
            self.r = deepcopy(self.initial_r)
        else:
            ## Expectation Propagation expectation value (prior).
            self.initial_r = deepcopy(r)
            ## Expectation Propagation expectation value (posterior).
            self.r = deepcopy(r)
        ## Expectation Propagation itemized covariance matrices.
        self.Q_features = []
        ## Expectation Propagation itemized expectation values.
        self.r_features = []
        if Q_features is None:
            for i in range(self.output_dim):
                self.Q_features.append(
                    deepcopy(self.Q) * 0
                )
        else:
            self.Q_features = deepcopy(Q_features)
        if r_features is None:
            for i in range(self.output_dim):
                self.r_features.append(
                    deepcopy(self.r) * 0
                )
        else:
            self.r_features = deepcopy(r_features)

        ## The inferred model parameters.
        self.inferred_parameters = {
            name: self.transform_parameters[name][0](self.final_expectation[i])
            for i, name in self.fixed_parameter_order
        }
        error_lower_bound = {}
        error_upper_bound = {}
        for i, name in self.fixed_parameter_order:
            error_lower_bound[name] = (
                self.final_expectation[i]
                - np.sqrt(
                    chi2(self.input_dim).ppf(0.95)
                ) * deviations[i])
            error_upper_bound[name] = (
                self.final_expectation[i]
                + np.sqrt(
                    chi2(self.input_dim).ppf(0.95)
                ) * deviations[i])
        error_lower_bound = {
            name: transform[0](error_lower_bound[name])
            for name, transform in self.transform_parameters.items()
        }
        error_upper_bound = {
            name: transform[0](error_upper_bound[name])
            for name, transform in self.transform_parameters.items()
        }
        ## The 95% confidence bounds (which don't reflect the
        # cross-correlations, but are easier to interpret).
        self.final_error_bounds = {
            name: (error_lower_bound[name], error_upper_bound[name])
            for name in self.log_of_tried_parameters.keys()
        }

    def result_to_json(self, seed=None):
        return json.dumps({
            "inferred parameters": self.inferred_parameters,
            "covariance":
                [list(line) for line in self.final_covariance],
            "correlation":
                [list(line) for line in self.final_correlation],
            "error bounds": self.final_error_bounds,
            "seed": seed,
            "Q": self.Q,
            "r": self.r,
            "Q_features": self.Q_features,
            "r_features": self.r_features,
            "fixed_parameter_order": self.fixed_parameter_order,
        }, cls=NDArrayEncoder)

    def log_to_json(self):
        return json.dumps({
            "tried parameters": self.log_of_tried_parameters,
            "discrepancies": self.log_of_discrepancies,
            "raw tried parameters": self.log_of_raw_tried_parameters,
        }, cls=NDArrayEncoder)

    def visualize_parameter_distribution(self):
        """!@brief Plots the features and visualizes the correlation.
        Please note that this function requires that the output of the
        individual simulators and the individual experimental data give
        an x- and y-axis when indexed with [0] and [1], respectively.
        Lists of lists in [0] and [1] with segmented data works as well.
        EP_BOLFI.run() does not have these restrictions.
        Visualizes the comparison that EP_BOLFI was set up to infer
        model parameters with. May be used to check if everything works
        as intended. Additionally, the 95% confidence error bounds for
        the parameters are visualized to check for reasonable bounds.
        If called after 'run()', the expected parameter set, the
        correlation and error bounds are from the finished estimation.
        """

        parameters = deepcopy(self.fixed_parameters)
        parameters.update({
            name: self.transform_parameters[name][0](value) for name, value
            in zip(self.log_of_tried_parameters.keys(), self.final_expectation)
        })
        combinations = combine_parameters_to_try(
            parameters, self.final_error_bounds
        )[0]
        fig_ax = [plt.subplots(figsize=(2**0.5 * 5, 5))
                  for _ in range(len(self.simulators))]
        for i, (sim, exp_data, feature) in enumerate(zip(
                self.simulators, self.experimental_datasets,
                self.feature_extractors)):
            ps = Preprocessed_Simulator(
                sim, self.fixed_parameters,
                list(self.log_of_tried_parameters.keys()),
                self.r, self.Q, exp_data, feature, self.transform_parameters,
                self.fixed_parameter_order
            )
            sim_data = ps.simulator(parameters)
            sim_data_plot = [[entry for segment in axis
                              for entry in np.atleast_1d(segment)]
                             for axis in sim_data]
            exp_data_plot = [[entry for segment in axis
                              for entry in np.atleast_1d(segment)]
                             for axis in exp_data]
            print("Features of simulator #" + str(i) + ":")
            print(feature(sim_data))
            print("Features of experiment #" + str(i) + ":")
            print(feature(exp_data))
            errorbars = [
                [entry
                 for segment in ps.simulator(combination)[1]
                 for entry in np.atleast_1d(segment)]
                for combination in combinations
            ]
            fig, ax = fig_ax[i]
            ax.plot(sim_data_plot[0], sim_data_plot[1],
                    label="expectation value")
            ax.plot(exp_data_plot[0], exp_data_plot[1],
                    label="experimental data")
            ax.fill_between(sim_data_plot[0],
                            np.min(errorbars, axis=0),
                            np.max(errorbars, axis=0), alpha=0.5,
                            label="95% confidence")
            ax.legend()
        plt.show()

    def run(
        self,
        bolfi_initial_evidence=None,
        bolfi_total_evidence=None,
        bolfi_posterior_samples=None,
        ep_iterations=3,
        ep_dampener=None,
        final_dampening=None,
        ep_dampener_reduction_steps=-1,
        gelman_rubin_threshold=None,
        ess_ratio_resample=5.0,
        ess_ratio_sampling_from_zero=-1.0,
        ess_ratio_abort=20.0,
        max_heuristic_steps=10,
        posterior_sampling_increase=1.2,
        model_resampling_increase=1.1,
        independent_mcmc_chains=4,
        scramble_ep_feature_order=True,
        show_trials=False,
        verbose=True,
        seed=None
    ):
        """!@brief Runs Expectation Propagation together with BOLFI.
        This function can be called multiple times; the estimation will
        take off from where it last stopped.
        @param bolfi_initial_evidence
            Number of evidence samples BOLFI will take for each feature
            before using Bayesian Optimization sampling. Default:
            1 + 2 ^ "number of estimated parameters".
        @param bolfi_total_evidence
            Number of evidence samples BOLFI will take for each feature
            in total (including initial evidence). Default:
            2 * "bolfi_initial_evidence".
        @param bolfi_posterior_samples
            Effective number of samples BOLFI will take from the
            posterior distribution. These are then used to fit a
            Gaussian to the posterior. Fit convergence scales with 1/√n.
            Default: I² + 3 * I with I as the number of
            estimated parameters. This is the number of the
            metaparameters of the underlying probability distribution
            times 2. The "times 2" considers the warmup samples.
        @param ep_iterations
            The number of iterations of the Expectation Propagation
            algorithm, i.e., the number of passes over each feature.
            Default: 3.
        @param ep_dampener
            The linear combination factor of the posterior calculated by
            BOLFI and the pseudo-prior. 0 means no dampening, i.e., the
            pseudo-prior gets replaced by the posterior. For values up
            to 1, that fraction of the pseudo-prior remains in each
            site update. Default: with "a" as the number of features and
            "b" as "ep_iterations", 1 - a * (1 - ᵃᵇ√"final_dampening").
        @param final_dampening
            Alternative way to set "ep_dampener". 0 means no dampening.
            For values up to 1, that fraction of the prior remains after
            the whole estimation. Default: if "ep_dampener" is not set,
            0.5. Else, "ep_dampener" takes precedence.
        @param ep_dampener_reduction_steps
            Number of iterations over which the 'ep_dampener' gets
            reduced to 0. In each iteration, an equal fraction of it
            gets subtracted. Set to a negative number to disable the
            reduction. Default: -1.
        @param gelman_rubin_threshold
            Optional threshold on top of the effective sample size.
            Values close to one indicate a converged estimate of the
            pseudo-posteriors. Never set to exactly one.
        @param ess_ratio_sampling_from_zero
            Threshold in the ratio of effective sample size to samples
            in the pseudo-posterior estimation, at which the sampling
            defaults to starting at the center of the pseudo-prior. Set
            higher than "ess_ratio_resample" to disable this behaviour.
        @param ess_ratio_resample
            Threshold in the ratio of effective sample size to samples
            in the pseudo-posterior estimation, at which before sampling
            the model gets resampled. Set higher than "ess_ratio_abort"
            to disable this behaviour.
        @param ess_ratio_abort
            Threshold in the ratio of effective sample size to samples
            in the pseudo-posterior estimation, at which the sampling
            aborts and the pseudo-posterior update is skipped.
        @param max_heuristic_steps
            The heuristics that are set by the "ess_ratio_*" arguments
            could effectively run forever. This parameter limits the
            amount of times these heuristics get employed in on EP
            iteration before it terminates.
        @param posterior_sampling_increase
            The factor by which the ratio of the effective sample size
            to samples in the pseudo-posterior estimation is multiplied
            each loop (cumulatively). Never set to exactly one or lower,
            as it might result in an infinite loop.
        @param model_resampling_increase
            The factor by which 'bolfi_total_evidence' gets multiplied
            each time the model gets resampled.
        @param independent_mcmc_chains
            The number of independent Markov-Chain Monte Carlo chains
            that are used for the estimation of the pseudo-posterior.
            Since we did not implement parallelization, more chains will
            not be faster, but more stable.
        @param scramble_ep_feature_order
            True randomizes the order that the EP features are iterated
            over. Their order is still consistent across EP iterations.
            False uses the order that the "feature_extractors" define.
        @param show_trials
            True plots the log of tried parameters live. Please note
            that each plot blocks the execution of the program, so do
            not use this when running the estimation in the background.
        @param verbose
            True shows verbose error messages and logs of the estimation
            process. With False, you need to get the estimation results
            from self.final_expectation and self.final_covariance.
            Default: True.
        @param seed
            Optional seed that is used in the RNG. If None is given, the
            results will be slightly different each time.
        @return
            The BOLFI instance of the last EP iteration. As such, it
            contains the Posterior of the overall inference procedure.
        """

        # Enable parallelization in ELFI. For details, see https://
        # elfi.readthedocs.io/en/latest/usage/parallelization.html
        # For local multithreading with all cores:
        # elfi.set_client('multiprocessing')
        # For scaling to clusters:
        # elfi.set_client('ipyparallel')

        if self.output_dim == 1 and ep_iterations > 1:
            raise RuntimeError(
                "EP with only one feature does not properly update the "
                "posterior after the initial loop. Set ep_iterations to 1. "
                "If you want to re-run, initialize a new EP_BOLFI instance "
                "from this one's Q and r attributes, but leave the attributes "
                "Q_features and r_features as None. "
                "This will make the former posterior the new prior."
            )

        bolfi_initial_evidence = (bolfi_initial_evidence or
                                  1 + 2**self.input_dim)
        bolfi_total_evidence = (bolfi_total_evidence or
                                2 * bolfi_initial_evidence)
        bolfi_posterior_samples = (
            bolfi_posterior_samples or
            self.input_dim**2 + 3 * self.input_dim
        )
        final_dampening = final_dampening or 0.5
        final_dampening = np.max([0, np.min([final_dampening, 1])])
        ep_dampener = (
            ep_dampener or
            1 - self.output_dim * (
                1 - final_dampening**(1 / (self.output_dim * ep_iterations))
            )
        )
        ep_dampener = np.max([0, np.min([ep_dampener, 1])])
        gelman_rubin_threshold = gelman_rubin_threshold or float('inf')
        gelman_rubin_threshold = (
            gelman_rubin_threshold if gelman_rubin_threshold > 1 else 1.1
        )
        posterior_sampling_increase = (
            posterior_sampling_increase if posterior_sampling_increase > 1.0
            else 1.2
        )
        model_resampling_increase = (
            model_resampling_increase if model_resampling_increase > 1.0
            else 1.2
        )

        # Shuffle the order in which the features are processed.
        if scramble_ep_feature_order:
            shuffled_order = (
                np.random.default_rng(seed).permutation(self.output_dim)
            )
        else:
            shuffled_order = np.array(range(self.output_dim))

        skip_sample = False

        for k in range(ep_iterations):

            # Compute the dampening for this iteration.
            if ep_dampener_reduction_steps > 0:
                if (k + 1 > ep_dampener_reduction_steps or
                        ep_dampener_reduction_steps == 0):
                    ep_dampener = 0.0
                else:
                    ep_dampener = ep_dampener * (
                        1.0 - k / ep_dampener_reduction_steps
                    )

            for j in shuffled_order:

                simulator_index = self.simulator_index_by_feature[j]
                sub_index = self.sub_index_by_feature[j]

                if verbose:
                    print()
                    if self.display_current_feature is not None:
                        print(
                            "Current feature: "
                            + self.display_current_feature[simulator_index](
                                sub_index
                            ) + "."
                        )
                    else:
                        print("Current feature: #" + str(j) + ".")

                # Initialize a new model in ELFI to ensure that no old,
                # superfluous nodes get added to this inference task.
                elfi.new_model()

                # Initialize the ELFI priors. For technical reasons, the
                # artificial "stabilising noise" used in BOLFI is equal
                # across all parameters. So in order to not have the
                # parameters with larger variance blow up the noise, all
                # should be scaled to have equal variance. To avoid any
                # accidental mismatch of means, they are normed to 0 and
                # transformed back into the actual parameter ranges when
                # needed.
                priors = []
                for i in range(self.input_dim):
                    priors.append(elfi.Prior("norm", 0.0, 1.0, name=str(i)))

                # Compute the pseudo priors.
                Q_pseudoprior = self.Q - self.Q_features[j]
                r_pseudoprior = self.r - self.r_features[j]

                # Get the simulator for use with ELFI with its
                # normalization and transformation functions.
                ps = Preprocessed_Simulator(
                    self.simulators[simulator_index], self.fixed_parameters,
                    list(self.log_of_tried_parameters.keys()), r_pseudoprior,
                    Q_pseudoprior, self.experimental_datasets[simulator_index],
                    self.feature_extractors[simulator_index],
                    self.transform_parameters,
                    self.fixed_parameter_order
                )

                # Create the ELFI simulator.
                elfi_simulator = elfi.Simulator(
                    ps.elfi_simulator, *priors,
                    observed=self.experimental_features[simulator_index]
                )

                # Create the summary operation that selects the feature.
                # Note that it normalizes to the experimental features.
                # This gives equal weight to every feature before
                # "weights" are applied.
                def summary(data):
                    try:
                        return [
                            data[sub_index] / self.experimental_features[
                                simulator_index
                            ][sub_index]
                            * self.weights[simulator_index][sub_index]
                        ]
                    except IndexError:
                        if verbose:
                            print("Warning: simulator output was too short.")
                        return [0]
                processed_data = elfi.Summary(summary, elfi_simulator)

                # Use the Euclidean distance. Either the feature is a
                # number and it is irrelevant, or it is a time series.
                distance = elfi.Distance('euclidean', processed_data)

                # It may be beneficial to take the logarithm of the
                # discrepancy. This is because the Gaussian Process used
                # in BOLFI is highly sensitive to high discrepancies
                # (which shouldn't be accounted for too much,
                # since they just correspond to bad parameter guesses).
                log_distance = elfi.Operation(np.log, distance)

                # Calculate the hard boundaries for BOLFI.
                # if self.free_parameters_boundaries is None:
                bolfi_bounds = {
                    str(c): (-np.sqrt(chi2(self.input_dim).ppf(0.95)),
                             +np.sqrt(chi2(self.input_dim).ppf(0.95)))
                    for c in range(self.input_dim)
                }
                """
                else:
                    bolfi_bounds = {}
                    lower_bounds = {k: v[0] for k, v
                                    in self.free_parameters_boundaries.items()}
                    upper_bounds = {k: v[1] for k, v
                                    in self.free_parameters_boundaries.items()}
                    search_lower_bounds = ps.transformed_trial_to_search(
                        ps.apply_transformation(lower_bounds)
                    )
                    search_upper_bounds = ps.transformed_trial_to_search(
                        ps.apply_transformation(upper_bounds)
                    )
                    for c in range(self.input_dim):
                        bolfi_bounds[str(c)] = (
                            search_lower_bounds[c], search_upper_bounds[c]
                        )
                    # Since the directions might be flipped, ensure that
                    # the lower bounds are lower than the upper bounds.
                    for key, value in bolfi_bounds.items():
                        if value[1] < value[0]:
                            bolfi_bounds[key] = (value[1], value[0])
                """

                # Start BOLFI and show a progressbar if requested.
                # The 'async' keyword can't be used since it is reserved
                # in newer Python versions.
                bolfi = elfi.BOLFI(
                    log_distance, initial_evidence=bolfi_initial_evidence,
                    batch_size=1, bounds=bolfi_bounds, seed=seed
                )
                bolfi.fit(n_evidence=bolfi_total_evidence, bar=verbose)

                state = Optimizer_State(
                    self.input_dim,
                    independent_mcmc_chains,
                    bolfi_total_evidence,
                    bolfi_posterior_samples,
                    gelman_rubin_threshold,
                    ess_ratio_resample,
                    ess_ratio_sampling_from_zero,
                    ess_ratio_abort,
                    posterior_sampling_increase,
                    model_resampling_increase,
                )

                degenerate_estimate = np.array([0.0] * self.input_dim)
                minimum_ess = 0
                ess = [0 for _ in range(self.input_dim)]
                maximum_gelman_rubin = float('inf')
                gelman_rubin = [float('inf') for _ in range(self.input_dim)]
                sample_ess_ratio = 1.0
                current_action = None
                adapt_sample_size = True
                heuristic_steps = 0

                while (minimum_ess < bolfi_posterior_samples or
                        (maximum_gelman_rubin > gelman_rubin_threshold if
                         gelman_rubin_threshold < float('inf') else False)):
                    state.calculate_next_step(sample_ess_ratio, current_action)
                    if state.current_action == "resample":
                        adapt_sample_size = False
                    current_action = None
                    if (
                        state.current_action == "abort"
                        or
                        heuristic_steps > max_heuristic_steps
                    ):
                        degenerate_estimate = np.zeros(self.input_dim)
                        skip_sample = True
                        print(state.verbose_actions["abort"])
                        break
                    if verbose:
                        # print(bolfi.target_model)
                        print(state.verbose_actions[state.current_action])
                    heuristic_steps += 1
                    # If the number of evidence points stayed the same,
                    # this just returns the current posterior.
                    bolfi.fit(n_evidence=state.total_evidence, bar=verbose)
                    try:
                        # Make a deepcopy of the BOLFI instance for sampling.
                        # Reason: when the sampling errors out with
                        # "NUTS: Cannot find acceptable stepsize ...",
                        # something breaks internally so that bolfi.fit
                        # errors out with inconsistent array sizes.
                        # File "<lib>/elfi/methods/bo/gpy_regression.py",
                        #   line 132,
                        #   in predict:
                        # r2 = np.sum(x**2., 1)[:, None]
                        #    + self._rbf_x2sum
                        #    - 2. * x.dot(self._gp.X.T)
                        # ValueError: operands could not be broadcast together
                        # with shapes (1,<old evidence>) (1,<old evidence> + 1)
                        bolfi_for_sampling = deepcopy(bolfi)
                        result = bolfi_for_sampling.sample(
                            int(
                                state.posterior_samples * sample_ess_ratio
                            ) if adapt_sample_size else int(
                                state.posterior_samples
                            ),
                            n_chains=independent_mcmc_chains,
                            initials=state.initials
                        )
                        adapt_sample_size = True
                    except OverflowError as oe:
                        # cannot convert float infinity to integer
                        # Happens when state.posterior_samples gets too large.
                        if verbose:
                            print("Error message from BOLFI:", oe)
                        skip_sample = True
                        break
                    except ValueError as ve:
                        # NUTS: Cannot find acceptable stepsize
                        # starting from point [best evidence].
                        if verbose:
                            print("Error message from BOLFI:", ve)
                        em = str(ve)
                        degenerate_estimate = np.fromstring(
                            em[em.find('[')+1:em.find(']')],
                            dtype=float, sep='  '
                        )
                        current_action = "resample"
                        continue
                    except ZeroDivisionError:
                        skip_sample = True
                        break
                    chains = result.meta['chains']
                    for i in range(self.input_dim):
                        ess[i] = (
                            elfi.methods.mcmc.eff_sample_size(chains[:, :, i])
                        )
                        gelman_rubin[i] = (
                            elfi.methods.mcmc.gelman_rubin_statistic(
                                chains[:, :, i]
                            )
                        )
                    minimum_ess = np.min(ess)
                    minimum_ess = 1 if np.min(ess) <= 0 else minimum_ess
                    # The samples_array excludes the warmup samples (one half).
                    # sample_ess_ratio is the amount of samples
                    # compared to the minimum ESS of any parameter. It's >= 1.
                    sample_ess_ratio = (
                        2 * len(result.samples_array) / minimum_ess
                    )
                    maximum_gelman_rubin = np.max(gelman_rubin)

                if self.display_current_feature is not None:
                    self.log_of_raw_tried_parameters[
                        self.display_current_feature[
                            self.simulator_index_by_feature[j]
                        ](self.sub_index_by_feature[j])
                    ].extend(bolfi.target_model.X)
                    self.log_of_discrepancies[
                        self.display_current_feature[
                            self.simulator_index_by_feature[j]
                        ](self.sub_index_by_feature[j])
                    ].extend(bolfi.target_model.Y)
                else:
                    self.log_of_raw_tried_parameters[str(j)].extend(
                        bolfi.target_model.X
                    )
                    self.log_of_discrepancies[str(j)].extend(
                        bolfi.target_model.Y
                    )

                # Append the newly tried parameters.
                for name in self.log_of_tried_parameters.keys():
                    self.log_of_tried_parameters[name].extend(
                        ps.log_of_tried_parameters[name]
                    )

                if show_trials:
                    # Plot the trial log.
                    fig, axes = plt.subplots(
                        self.input_dim, 1, num=0, clear=True,
                        constrained_layout=True
                    )
                    axes = np.atleast_1d(axes)
                    for index, name in zip(
                        range(self.input_dim),
                        self.log_of_tried_parameters.keys()
                    ):
                        axes[index].set_title(name)
                        axes[index].semilogy(range(len(
                            self.log_of_tried_parameters[name]
                        )), self.log_of_tried_parameters[name],
                            marker='.', lw=0, label=name
                        )
                    plt.show()

                if skip_sample:
                    try:
                        inferred_parameters = ps.undo_transformation(
                            ps.search_to_transformed_trial(
                                degenerate_estimate
                            )
                        )
                        if verbose:
                            print("inferred parameters:")
                            print(inferred_parameters)
                    except (ValueError, IndexError):
                        # BOLFI probably didn't take any samples at all.
                        pass
                    skip_sample = False
                    continue

                # if verbose:
                #     print(result)
                #     print(result.samples_array)

                # Get the raw samples. They are given as a list.
                normed_samples_0 = np.array(result.samples_array)

                # Un-raw the samples.
                samples = ps.search_to_transformed_trial(normed_samples_0)
                samples = [samples[name] for name
                           in self.log_of_tried_parameters.keys()]
                number_of_samples = len(samples[0])

                # Compute mean and covariance directly from the samples.
                µ_h = np.sum(samples, axis=1) / number_of_samples
                # np.outer(vector) computes vector.T * vector.
                covariance = np.sum(
                    np.apply_along_axis(
                        lambda vector: np.outer(vector, vector), 1,
                        np.apply_along_axis(
                            lambda slice: slice - µ_h, 0, samples
                        ).T
                    ),
                    axis=0
                ) / (number_of_samples - 1)

                # Calculate the Expectation Propagation site update.
                Q_h = np.linalg.inv(covariance)
                r_h = Q_h @ µ_h

                self.Q_features[j] += (1.0 - ep_dampener) * (Q_h - self.Q)
                self.r_features[j] += (1.0 - ep_dampener) * (r_h - self.r)

                self.Q = (1.0 - ep_dampener) * Q_h + ep_dampener * self.Q
                self.r = (1.0 - ep_dampener) * r_h + ep_dampener * self.r

                # Update the covariance and expectation values.
                self.final_covariance = np.linalg.inv(self.Q)
                self.final_expectation = self.final_covariance @ self.r

                self.inferred_parameters = ps.undo_transformation(
                    {
                        name: self.final_expectation[i]
                        for i, name in self.fixed_parameter_order
                    }
                )

                # Calculate the correlation matrix for interpretability.
                deviations = np.sqrt(np.diag(self.final_covariance))
                # 'None' in this context is equivalent to 'np.newaxis'.
                self.final_correlation = (
                    self.final_covariance / deviations[:, None]
                ) / deviations[None, :]

                # Calculate the error bounds (which don't reflect the
                # correlations, but are easier to interpret).
                error_lower_bound = {}
                error_upper_bound = {}
                for i, name in self.fixed_parameter_order:
                    error_lower_bound[name] = (
                        self.final_expectation[i]
                        - np.sqrt(
                            chi2(self.input_dim).ppf(0.95)
                        ) * deviations[i])
                    error_upper_bound[name] = (
                        self.final_expectation[i]
                        + np.sqrt(
                            chi2(self.input_dim).ppf(0.95)
                        ) * deviations[i])
                error_lower_bound = ps.undo_transformation(error_lower_bound)
                error_upper_bound = ps.undo_transformation(error_upper_bound)
                self.final_error_bounds = {
                    name: (error_lower_bound[name], error_upper_bound[name])
                    for name in self.log_of_tried_parameters.keys()
                }

                if verbose:
                    print("inferred parameters:", self.inferred_parameters)
                    print("covariance of inferred parameters:")
                    print(repr(self.final_covariance))
                    print("correlation of inferred parameters:")
                    print(repr(self.final_correlation))
                    print("resulting approximate error bounds:")
                    print("{")
                    for name, value in self.final_error_bounds.items():
                        print("'" + name + "':", value, ",")
                    print("}")

        if verbose:
            # Additionally display the posteriors per feature.
            print()
            print("final state of the estimator:")
            print("inferred parameters:", self.inferred_parameters)
            print("covariance of inferred parameters:")
            print(repr(self.final_covariance))
            print("correlation of inferred parameters:")
            print(repr(self.final_correlation))
            print("resulting approximate error bounds:")
            print("{")
            for name, value in self.final_error_bounds.items():
                print("'" + name + "':", value, ",")
            print("}")
            print("precision matrix:")
            print("Q:")
            print(repr(self.Q))
            print("precision-scaled raw parameter guess:")
            print("r:")
            print(repr(self.r))
            print("itemized precisions per feature:")
            print("Q_features:")
            print(repr(self.Q_features))
            print("itemized precision-scaled raw parameter guesses:")
            print("r_features:")
            print(repr(self.r_features))
            for i in range(self.output_dim):
                print()
                if self.display_current_feature is not None:
                    simulator_index = self.simulator_index_by_feature[i]
                    sub_index = self.sub_index_by_feature[i]
                    print(self.display_current_feature[simulator_index](
                        sub_index
                    ) + ":")
                else:
                    print("Feature #" + str(i) + ":")
                try:
                    local_covariance = np.linalg.inv(self.Q_features[i])
                except np.linalg.LinAlgError:
                    local_covariance = np.diag([
                        np.inf for i in range(self.input_dim)
                    ])
                print("covariance:")
                print(repr(local_covariance))
                # Calculate the correlation matrix for interpretability.
                local_deviations = np.sqrt(np.diag(local_covariance))
                # 'None' in this context is equivalent to 'np.newaxis'.
                local_correlation = (
                    local_covariance / local_deviations[:, None]
                ) / local_deviations[None, :]
                print("correlation:")
                print(repr(local_correlation))
                local_guess = local_covariance @ self.r_features[i]
                # Calculate the error bounds (which don't reflect the
                # correlations, but are easier to interpret).
                local_error_lower_bound = {}
                local_error_upper_bound = {}
                for i, name in self.fixed_parameter_order:
                    local_error_lower_bound[name] = (
                        local_guess[i] - np.sqrt(
                            chi2(self.input_dim).ppf(0.95)
                        ) * local_deviations[i]
                    )
                    local_error_upper_bound[name] = (
                        local_guess[i] + np.sqrt(
                            chi2(self.input_dim).ppf(0.95)
                        ) * local_deviations[i]
                    )
                local_error_lower_bound = ps.undo_transformation(
                    local_error_lower_bound)
                local_error_upper_bound = ps.undo_transformation(
                    local_error_upper_bound)
                local_error_bounds = {
                    name: (local_error_lower_bound[name],
                           local_error_upper_bound[name])
                    for name in self.log_of_tried_parameters.keys()
                }
                local_guess = ps.undo_transformation({
                    name: local_guess[i]
                    for i, name in self.fixed_parameter_order
                })
                print("guess:")
                for name, value in local_guess.items():
                    print(name, value)
                print("resulting approximate error bounds:")
                print("{")
                for name, value in local_error_bounds.items():
                    print("'" + name + "':", value, ",")
                print("}")

        # Return the last BOLFI instance, as it contains the Posterior.
        return bolfi



"""!@package ep_bolfi.utility.preprocessing
Contains frequently used workflows in dataset preprocessing.

The functions herein are a collection of simple, but frequent,
transformations of arrays of raw measurement data.
"""

import numpy as np
import copy
from collections.abc import MutableMapping
from contextlib import contextmanager
from itertools import chain
from multiprocessing import Pool
from numbers import Number
from pybamm import Scalar
from scipy.optimize import root_scalar
from scipy.stats import chi2, norm
from ep_bolfi.models.solversetup import solver_setup, simulation_setup
from ep_bolfi.utility.fitting_functions import (
    fit_OCV, inverse_OCV_fit_function, smooth_fit,
    verbose_spline_parameterization
)


class SubstitutionDict(MutableMapping):
    """!@brief A dictionary with some automatic substitutions.
    "substitutions" is a dictionary that extends "storage" with
    automatic substitution rules depending on its value types:
     - string, which serves the value of "storage" at that value.
     - callable which takes one parameter, which will get passed its
       SubstitutionDict instance and serves its return value.
     - any other type, which serves the value as-is.
    Assigning values to keys afterwards will overwrite substitutions.
    """

    def __init__(self, storage, substitutions={}):
        self._storage = storage
        self._substitutions = substitutions
        self._log = []
        self._log_switch = False

    def __delitem__(self, key):
        if key in self._storage.keys():
            self._storage.__delitem__(key)
        if key in self._substitutions.keys():
            self._substitutions.__delitem__(key)
        if (
            key not in self._storage.keys()
            and
            key not in self._substitutions.keys()
        ):
            raise KeyError(key)

    def __getitem__(self, key):
        if self._log_switch:
            self._log.append(key)
        if key in self._substitutions.keys():
            storage_value = self._substitutions[key]
            if isinstance(storage_value, Number):
                return storage_value
            elif isinstance(storage_value, str):
                return self[storage_value]
            elif callable(storage_value):
                return storage_value(self)
            else:
                # fallback
                return storage_value
        elif key in self._storage.keys():
            return self._storage[key]
        else:
            raise KeyError(key)

    def __iter__(self):
        return chain(self._storage, self._substitutions)

    def __len__(self):
        return len(set(
            list(self._storage.keys()) + list(self._substitutions.keys())
        ))

    def __setitem__(self, key, value):
        if key in self._substitutions.keys():
            del self._substitutions[key]
        self._storage[key] = value

    def __str__(self):
        return '{' + ', '.join(
            [str(k) + ': ' + str(v) for k, v in self.items()]
        ) + '}'

    def __repr__(self):
        return self.__str__()

    @contextmanager
    def log_lock(self):
        self._log_switch = True
        self._log = []
        try:
            yield
        finally:
            self._log_switch = False
            self._log = []

    def dependent_variables(self, parameters):
        with self.log_lock():
            for p in parameters:
                self[p]
            log = copy.deepcopy(self._log)
        for s in self._substitutions.keys():
            with self.log_lock():
                self[s]
                for p in parameters:
                    if p in self._log:
                        log.append(s)
                        break
        return log


def fix_parameters(parameters_to_be_fixed):
    """!@brief Returns a function which sets some parameters in advance.
    @param parameters_to_be_fixed
        These parameters will at least be a
        part of the dictionary that the returned function returns.
    @return
        The function which adds additional parameters to a
        dictionary or replaces existing parameters with the new ones.
    """

    def return_all_parameters(free_parameters):
        """!@brief Adds the 'free_parameters' to a pool of parameters.
        @param free_parameters
            A dictionary which gets added to the fixed pool of
            parameters (see the function 'fix_parameters').
        @return
            A dictionary containing free_parameters and some other
            key-value-pairs as defined by 'fix_parameters'. If a key is
            present in both, the value from 'free_parameters' is used.
        """

        return_dict = copy.deepcopy(parameters_to_be_fixed)
        return_dict.update(free_parameters)
        return return_dict

    return return_all_parameters


def combine_parameters_to_try(parameters, parameters_to_try_dict):
    """!@brief Give every combination as full parameter sets.
    Compatible with SubstitutionDict, if "parameters" is one.
    @param parameters
        The base full parameter set as a dictionary.
    @param parameters_to_try_dict
        The keys of this dictionary correspond to the "parameters"' keys
        where different values are to be inserted. These are given by
        the tuples which are the values of this dictionary.
    @return
        A 2-tuple where the first item is the list of all parameter set
        combinations and the second the list of the combinations only.
    """

    # Fetch the parameters to try.
    parameters_to_try = []
    for key, value in parameters_to_try_dict.items():
        parameters_to_try.append((key, value))

    def recursive_combination(last_index):
        """! Recursively get every combination of parameters to try. """
        return_list = []
        if last_index == 0:
            key = parameters_to_try[0][0]
            for value in parameters_to_try[0][1]:
                return_list.append({key: value})
        elif last_index > 0:
            key = parameters_to_try[last_index][0]
            list_of_combinations = recursive_combination(last_index-1)
            for value in parameters_to_try[last_index][1]:
                for combination in list_of_combinations:
                    return_dict = {key: value}
                    return_dict.update(combination)
                    return_list.append(return_dict)
        return return_list

    combinations = recursive_combination(len(parameters_to_try)-1)
    parameters_list = []
    for combination in combinations:
        parameters_list.append(fix_parameters(parameters)(combination))
    return (parameters_list, combinations)


def calculate_means_and_standard_deviations(
    mean,
    covariance,
    free_parameters_names,
    transform_parameters={},
    bounds_in_standard_deviations=1,
    **kwargs
):
    """!@brief Calculate means and standard deviations.
    Please note that standard deviations translate differently into
    confidence regions in different dimensions. For the confidence
    region, use "approximate_confidence_ellipsoid".
    @param mean
        The mean of the uncertain parameters as a dictionary.
    @param covariance
        The covariance of the uncertain parameters as a two-dimensional
        numpy array.
    @param free_parameters_names
        The names of the parameters that are uncertain as a list. This
        parameter maps the order of parameters in "covariance".
    @param transform_parameters
        Optional transformations between the parameter space that is
        used for searching for optimal parameters and the model
        parameters. Any missing free parameter is not transformed.
        The values are 2-tuples. The first entry is a function taking
        the search space parameter and returning the model parameter.
        The second entry is the inverse function.
        For convenience, any value may also be one of the following:
         - 'none' => (identity, identity)
         - 'log' => (exp, log)
    @param bounds_in_standard_deviations
        Sets how many standard deviations in each direction the returned
        error bounds are. These are first applied and then transformed.
    @param **kwargs
        Keyword arguments for scipy.integrate.quad, which is used to
        numerically calculate mean and variance.
    @return
        A 3-tuple with three dictionaries. Their keys are the free
        parameters' names as keys and their values are those parameters'
        means, standard deviations, and error bounds.
    """

    # Substitute transformations given by name.
    if transform_parameters is not {}:
        for name, function in transform_parameters.items():
            if type(function) is str:
                if function == 'none':
                    transform_parameters[name] = (
                        lambda s: s, lambda b: b
                    )
                elif function == 'log':
                    transform_parameters[name] = (
                        lambda s: np.exp(s), lambda b: np.log(b)
                    )
    # Fill-in 'none' transformations.
    for name in free_parameters_names:
        if name not in transform_parameters.keys():
            transform_parameters[name] = (
                lambda s: s, lambda b: b
            )

    means = {}
    standard_deviations = {}
    error_bounds = {}

    for i, name in enumerate(free_parameters_names):
        mean_internal = transform_parameters[name][1](mean[name])
        variance_internal = covariance[i][i]
        mean_actual = norm.expect(
            lambda x: (transform_parameters[name][0](
                mean_internal + x * np.sqrt(variance_internal)
            )),
            **kwargs
        )
        variance_actual = norm.expect(
            lambda x: (transform_parameters[name][0](
                mean_internal + x * np.sqrt(variance_internal)
            ))**2,
            **kwargs
        ) - mean_actual**2
        means[name] = mean_actual
        standard_deviations[name] = np.sqrt(variance_actual)
        error_bounds[name] = (
            transform_parameters[name][0](
                mean_internal
                - bounds_in_standard_deviations * np.sqrt(variance_internal)
            ),
            transform_parameters[name][0](
                mean_internal
                + bounds_in_standard_deviations * np.sqrt(variance_internal)
            ),
        )

    return (means, standard_deviations, error_bounds)


def approximate_confidence_ellipsoid(
    parameters,
    free_parameters_names,
    covariance,
    mean=None,
    transform_parameters={},
    refinement=True,
    confidence=0.95
):
    """!@brief Approximate a confidence ellipsoid.
    Compatible with SubstitutionDict, if "parameters" is one.
    The geometric approximation is a refinement of the polytope with
    nodes on the semiaxes of the confidence ellipsoid. The refinement
    step adds a node for each face, i.e., each sub-polytope with
    dimension smaller by 1. This node is centered on that face and
    projected onto the confidence ellipsoid.
    @param parameters
        The base full parameter set as a dictionary.
    @param free_parameters_names
        The names of the parameters that are uncertain as a list. This
        parameter has to match the order of parameters in "covariance".
    @param covariance
        The covariance of the uncertain parameters as a two-dimensional
        numpy array.
    @param mean
        The mean of the uncertain parameters as a dictionary. If not
        set, the values from 'parameters' will be used.
    @param transform_parameters
        Optional transformations between the parameter space that is
        used for searching for optimal parameters and the model
        parameters. Any missing free parameter is not transformed.
        The values are 2-tuples. The first entry is a function taking
        the search space parameter and returning the model parameter.
        The second entry is the inverse function.
        For convenience, any value may also be one of the following:
         - 'none' => (identity, identity)
         - 'log' => (exp, log)
    @param confidence
        The confidence within the ellipsoid. Defaults to 0.95, i.e., the
        95% confidence ellipsoid.
    @param refinement
        If False, only the nodes on the semiaxes get returned. If True,
        the nodes centered on the faces get returned as well.
    @return
        A 2-tuple where the first item is the list of all parameter set
        combinations and the second the ellipsoid nodes only as a
        two-dimensional numpy array with each node in on row.
    """

    if mean is None:
        mean = parameters
    # Substitute transformations given by name.
    if transform_parameters is not {}:
        for name, function in transform_parameters.items():
            if type(function) is str:
                if function == 'none':
                    transform_parameters[name] = (
                        lambda s: s, lambda b: b
                    )
                elif function == 'log':
                    transform_parameters[name] = (
                        lambda s: np.exp(s), lambda b: np.log(b)
                    )
    # Fill-in 'none' transformations.
    for name in free_parameters_names:
        if name not in transform_parameters.keys():
            transform_parameters[name] = (
                lambda s: s, lambda b: b
            )

    mahalanobis_squared = chi2(len(free_parameters_names))
    standard_deviations = np.sqrt(mahalanobis_squared.ppf(confidence))

    semiaxes_length_squared, semiaxes_normed = np.linalg.eigh(covariance)
    semiaxes_length = np.sqrt(semiaxes_length_squared)
    confidence_semiaxes_length = semiaxes_length * standard_deviations

    # These calculations happen in the eigenspace of the covariance.
    semiaxis_nodes = []
    for dim, csl in enumerate(confidence_semiaxes_length):
        semiaxis_nodes.append(
            [0.0] * dim
            + [-csl]
            + [0.0] * (len(free_parameters_names) - 1 - dim)
        )
        semiaxis_nodes.append(
            [0.0] * dim
            + [csl]
            + [0.0] * (len(free_parameters_names) - 1 - dim)
        )

    if refinement:
        # [aᵢ / √n]ᵢⁿ fulfills ∑ᵢⁿ (aᵢ / √n)² / aᵢ² = 1.
        diagonal = np.array(
            [-confidence_semiaxes_length,
             confidence_semiaxes_length]
        ) / np.sqrt(len(free_parameters_names))

        def recursive_combination(length, index=0):
            """! Recursively get every combination of nodes. """
            return_list = []
            if index == length - 1:
                for entry in diagonal.T[index]:
                    return_list.append([entry])
            elif index < length - 1:
                list_of_combinations = recursive_combination(length, index + 1)
                for entry in diagonal.T[index]:
                    for combination in list_of_combinations:
                        node = [entry]
                        node.extend(combination)
                        return_list.append(node)
            return return_list

        combinations = recursive_combination(len(free_parameters_names))
    else:
        combinations = []

    transformed_mean = np.array([
        transform_parameters[name][1](mean[name])
        for name in free_parameters_names
    ])
    transformed_ellipsoid = transformed_mean + np.array([
        (semiaxes_normed @ node)
        for node in [*semiaxis_nodes, *combinations]
    ])
    ellipsoid = np.array([
        [
            transform_parameters[name][0](v[i])
            for i, name in enumerate(free_parameters_names)
        ]
        for v in transformed_ellipsoid
    ])
    parameters_list = []
    for node in ellipsoid:
        node_parameters = {
            name: node[i] for i, name in enumerate(free_parameters_names)
        }
        parameters_list.append(fix_parameters(parameters)(node_parameters))
    return (parameters_list, ellipsoid)


def capacity(parameters, electrode="positive"):
    """!@brief Convenience function for calculating the capacity.
    @param parameters
        A parameter file as defined by models.standard_parameters.
    @param electrode
        The prefix of the electrode to use for capacity calculation.
        Change to "negative" to use the one with the lower OCP.
    @return
        The capacity of the parameterized battery in C.
    """

    return (
        parameters[
            electrode.capitalize()
            + " electrode active material volume fraction"
        ]
        * parameters[electrode.capitalize() + " electrode thickness [m]"]
        * parameters["Current collector perpendicular area [m2]"]
        * parameters[
            "Maximum concentration in " + electrode + " electrode [mol.m-3]"
        ]
        * 96485.33212
    )


def calculate_SOC(timepoints, currents, initial_SOC=0, sign=1, capacity=1):
    """!@brief Transforms applied current over time into SOC.
    @param timepoints
        Array of the timepoint segments.
    @param currents
        Array of the current segments.
    @param initial_SOC
        The SOC value to start accumulating from.
    @param sign
        The value by which to multiply the current.
    @param capacity
        A scaling by which to convert from C to dimensionless SOC.
    @return
        An array of the same shape describing SOC in C.
    """

    complete_SOC = []
    for t, I in zip(timepoints, currents):
        Δt = np.array([t1 - t0 for (t0, t1) in zip(t[:-1], t[1:])])
        try:
            I_int = np.array([0.5 * (I1 + I0)
                             for (I0, I1) in zip(I[:-1], I[1:])])
        except TypeError:
            I_int = np.atleast_1d(I)
        complete_SOC.append(np.append(
            initial_SOC,
            initial_SOC + sign * np.cumsum(I_int * Δt) / capacity
        ))
        initial_SOC = complete_SOC[-1][-1]
    return complete_SOC


def calculate_both_SOC_from_OCV(
    parameters,
    negative_SOC_from_cell_SOC,
    positive_SOC_from_cell_SOC,
    OCV
):
    """!@brief Calculates the SOC of both electrodes from their OCV.
    The SOCs are substitued in the given "parameters". The SOC of the
    cell as a whole gets returned in case it is needed.
    @param parameters
        The parameters of the battery as used for the PyBaMM simulations
        (see models.standard_parameters).
    @param negative_SOC_from_cell_SOC
        A function that takes the SOC of the cell and returns the SOC of
        the negative electrode.
    @param positive_SOC_from_cell_SOC
        A function that takes the SOC of the cell and returns the SOC of
        the positive electrode.
    @param OCV
        The OCV for which the SOCs shall be calculated.
    @return
        The SOC of the cell as a whole.
    """

    positive_OCV = parameters["Positive electrode OCP [V]"]
    negative_OCV = parameters["Negative electrode OCP [V]"]
    cell_SOC = root_scalar(
        lambda cell_SOC: positive_OCV(positive_SOC_from_cell_SOC(cell_SOC))
        - negative_OCV(negative_SOC_from_cell_SOC(cell_SOC)) - OCV,
        method='toms748', bracket=[0, 1], x0=0.5
    ).root
    positive_SOC = positive_SOC_from_cell_SOC(cell_SOC)
    negative_SOC = negative_SOC_from_cell_SOC(cell_SOC)
    parameters["Initial concentration in positive electrode [mol.m-3]"] = (
        positive_SOC
        * parameters["Maximum concentration in positive electrode [mol.m-3]"]
    )
    parameters["Initial concentration in negative electrode [mol.m-3]"] = (
        negative_SOC
        * parameters["Maximum concentration in negative electrode [mol.m-3]"]
    )

    return cell_SOC


def subtract_OCV_curve_from_cycles(
    dataset,
    parameters,
    starting_SOC=None,
    starting_OCV=None,
    electrode="positive",
    current_sign=0,
    voltage_sign=0,
):
    """!@brief Removes the OCV curve from a cycling measurement.
    @param dataset
        A Cycling_Information object of the measurement.
    @param parameters
        The parameters of the battery as used for the PyBaMM simulations
        (see models.standard_parameters).
    @param starting_SOC
        The SOC at the beginning of the measurement. If not given, the
        OCV curve will be inverted to determine the initial SOC.
    @param starting_OCV
        The OCV at the beginning of the measurement. If not given and
        starting_SOC is also not given, the first entry of voltages is
        used for this. If not given, but starting_SOC is, the OCP
        function will be evaluated at starting_SOC to get the OCV.
    @param electrode
        "positive" (default) or "negative" for current sign correction and
        capacity calculation. "positive" adds SOC with positive current and
        vice versa. The sign corrections can be overwritten with '*_sign'.
    @param current_sign
        1 adds SOC, -1 subtracts it, 0 follows the default behaviour above.
    @param voltage_sign
        1 subtracts the OCP, -1 adds it, 0 follows the default behaviour above.
    @return
        2-tuple. First entry are the voltages minus the OCV as estimated for
        each data point. These are structured in exactly the same way as in the
        "dataset". Second entry are the electrode SOCs as counted in the data.
    """

    if current_sign == 0:
        if electrode == "positive":
            current_sign = 1
        else:
            current_sign = -1
    if voltage_sign == 0:
        if electrode == "positive":
            voltage_sign = 1
        else:
            voltage_sign = -1

    OCV_function = parameters[electrode.capitalize() + " electrode OCP [V]"]
    if type(OCV_function(0.5)) is Scalar:
        def OCV_function(s): return (
            parameters[electrode.capitalize() + " electrode OCP [V]"](s).value
        )
    if starting_SOC is None:
        starting_OCV = starting_OCV or dataset.voltages[0][0]
        initial_SOC = root_scalar(
            lambda s: OCV_function(s) - starting_OCV,
            method='toms748',
            bracket=[0, 1],
            x0=0.5
        ).root
    else:
        initial_SOC = starting_SOC
        starting_OCV = starting_OCV or OCV_function(initial_SOC)
    C = capacity(parameters, electrode)
    SOCs = calculate_SOC(
        dataset.timepoints,
        dataset.currents,
        initial_SOC=initial_SOC,
        sign=current_sign,
        capacity=C,
    )

    returned_SOCs = copy.deepcopy(SOCs)

    # Skip over segments where nothing happens anyway.
    SOC_skip_indices = []
    for i, current in enumerate(dataset.currents):
        if not any(current):  # Read "if current is perfectly 0".
            SOC_skip_indices.append(i)
    for i in reversed(SOC_skip_indices):
        SOCs.pop(i)
    open_circuit_voltages = []
    for soc in SOCs:
        open_circuit_voltages.append(list(OCV_function(soc)))
    if 0 in SOC_skip_indices:
        open_circuit_voltages.insert(
            0, [starting_OCV] * len(dataset.timepoints[0])
        )
        SOC_skip_indices.pop(0)
    for i in SOC_skip_indices:
        open_circuit_voltages.insert(
            i, [open_circuit_voltages[i - 1][-1]] * len(dataset.timepoints[i])
        )
    overpotentials = [
        [
            U_entry - voltage_sign * OCV_entry
            for U_entry, OCV_entry in zip(U, OCV)
        ]
        for U, OCV in zip(dataset.voltages, open_circuit_voltages)
    ]
    return overpotentials, returned_SOCs


def subtract_both_OCV_curves_from_cycles(
    dataset,
    parameters,
    negative_SOC_from_cell_SOC,
    positive_SOC_from_cell_SOC,
    starting_SOC=None,
    starting_OCV=None,
):
    """!@brief Removes the OCV curve from a single cycle.
    @param dataset
        A Cycling_Information object of the measurement.
    @param parameters
        The parameters of the battery as used for the PyBaMM simulations
        (see models.standard_parameters).
    @param negative_SOC_from_cell_SOC
        A function that takes the SOC of the cell and returns the SOC of
        the negative electrode.
    @param positive_SOC_from_cell_SOC
        A function that takes the SOC of the cell and returns the SOC of
        the positive electrode.
    @param starting_SOC
        The SOC at the beginning of the measurement. If not given, the
        OCV curves will be inverted to determine the initial SOC.
    @param starting_OCV
        The OCV at the beginning of the measurement. If not given, the
        first entry of voltages is used for this.
    @return
        2-tuple. First entry are the voltages minus the OCV as estimated for
        each data point. These are structured in exactly the same way as in the
        "dataset". Second entry are the electrode SOCs as counted in the data.
    """

    positive_OCV = parameters["Positive electrode OCP [V]"]
    if type(positive_OCV(0.5)) is Scalar:
        def positive_OCV(s): return (
            parameters["Positive electrode OCP [V]"](s).value
        )
    negative_OCV = parameters["Negative electrode OCP [V]"]
    if type(negative_OCV(0.5)) is Scalar:
        def negative_OCV(s): return (
            parameters["Negative electrode OCP [V]"](s).value
        )
    if starting_SOC is None:
        starting_OCV = starting_OCV or dataset.voltages[0][0]
        cell_SOC_tracker = root_scalar(
            lambda cell_SOC: (
                positive_OCV(positive_SOC_from_cell_SOC(cell_SOC))
                - negative_OCV(negative_SOC_from_cell_SOC(cell_SOC))
                - starting_OCV
            ),
            method='toms748',
            bracket=[0, 1],
            x0=0.5
        ).root
    else:
        cell_SOC_tracker = starting_SOC
        starting_OCV = starting_OCV or (
            positive_OCV(positive_SOC_from_cell_SOC(cell_SOC_tracker))
            - negative_OCV(negative_SOC_from_cell_SOC(cell_SOC_tracker))
        )
    C = capacity(parameters)
    returned_SOCs = []
    voltages = []
    for t, I, U in zip(dataset.timepoints, dataset.currents, dataset.voltages):
        Δt = [t1 - t0 for (t0, t1) in zip(t[:-1], t[1:])]
        try:
            I_int = [0.5 * (I0 + I1) for (I0, I1) in zip(I[:-1], I[1:])]
        except TypeError:
            I_int = np.atleast_1d(I)
        cell_SOC = cell_SOC_tracker + np.cumsum(
            np.array(I_int) * np.array(Δt)
        ) / C
        voltages.append([U[0] - starting_OCV] + list(
            np.array(U[1:])
            - np.array(positive_OCV(positive_SOC_from_cell_SOC(cell_SOC)))
            + np.array(negative_OCV(negative_SOC_from_cell_SOC(cell_SOC)))
        ))
        returned_SOCs.append(cell_SOC)
        cell_SOC_tracker = cell_SOC[-1]
        starting_OCV = (
            positive_OCV(positive_SOC_from_cell_SOC(cell_SOC_tracker))
            - negative_OCV(negative_SOC_from_cell_SOC(cell_SOC_tracker))
        )
    return voltages, returned_SOCs


def laplace_transform(x, y, s):
    """!@brief Performs a basic laplace transformation.
    @param x
        The independent variable.
    @param y
        The dependent variable.
    @param s
        The (possibly complex) frequencies for which to perform the
        transform.
    @return
        The evaluation of the laplace transform at s.
    """

    x = np.array(x)
    y = np.array(y)
    Δx = x[1:] - x[:-1]
    x_int = 0.5 * (x[1:] + x[:-1])
    y_int = 0.5 * (y[1:] + y[:-1])
    return np.sum(y_int * np.exp(-(s[:, None] + 0.0j) * x_int) * Δx, axis=1)


def find_occurrences(sequence, value):
    """!@brief Gives indices in sequence where it is closest to value.
    @param sequence
        A list that represents a differentiable function.
    @param value
        The value that is searched for in "sequence". Also, crossings of
        consecutive values in "sequence" with "value" are searched for.
    @return
        A list of indices in "sequence" in ascending order where "value"
        or a close match for "value" was found.
    """
    root_finder = np.array(sequence) - value
    crossings = (
        np.sign(root_finder[1:]) - np.sign(root_finder[:-1])
    ).nonzero()[0]
    nearest_indices = np.where(
        np.abs(root_finder[crossings]) < np.abs(root_finder[crossings + 1]),
        crossings, crossings + 1
    )
    if len(nearest_indices) == 0:
        nearest_indices = (
            [0]
            if np.abs(root_finder[0]) < np.abs(root_finder[-1]) else
            [len(sequence) - 1]
        )
    return nearest_indices


def OCV_from_CC_CV(
    charge,
    cv,
    discharge,
    name,
    phases,
    eval_points=200,
    spline_SOC_range=(0.01, 0.99),
    spline_order=2,
    spline_smoothing=2e-3,
    spline_print=None,
    parameters_print=False
):
    """!@brief Tries to extract the OCV curve from CC-CV cycling data.
    @param charge
        A Cycling_Information object containing the constant charge
        cycle(s). If more than one CC-CV-cycle shall be analyzed, please
        make sure that the order of this, cv and discharge align.
    @param cv
        A Cycling_Information object containing the constant voltage
        part between charge and discharge cycle(s).
    @param discharge
        A Cycling_Information object containing the constant discharge
        cycle(s). These occur after each cv cycle.
    @param name
        Name of the material for which the CC-CV-cycling was measured.
    @param phases
        Number of phases in the fitting_functions.OCV_fit_function as an
        int. The higher it is, the more (over-)fitted the model becomes.
    @param eval_points
        The number of points for plotting of the OCV curves.
    @param spline_SOC_range
        2-tuple giving the SOC range in which the inverted
        fitting_functions.OCV_fit_function will be interpolated by a
        smoothing spline. Outside of this range the spline is used for
        extrapolation. Use this to fit the SOC range of interest more
        precisely, since a fit of the whole range usually fails due to
        the singularities at SOC 0 and 1. Please note that this range
        considers the 0-1-range in which the given SOC lies and not the
        linear transformation of it from the fitting process.
    @param spline_order
        Order of this smoothing spline. If it is set to 0, only the
        fitting_functions.OCV_fit_function is calculated and plotted.
    @param spline_smoothing
        Smoothing factor for this smoothing spline. Default: 2e-3. Lower
        numbers give more precision, while higher numbers give a simpler
        spline that smoothes over steep steps in the fitted OCV curve.
    @param spline_print
        If set to either 'python' or 'matlab', a string representation
        of the smoothing spline is printed in the respective format.
    @param parameters_print
        Set to True if the fit parameters should be printed to console.
    @return
        A 8-tuple consisting of the following:
        0: OCV_fits
            The fitted OCV curve parameters for each CC-CV cycle as
            returned by fitting_functions.fit_OCV.
        1: I_mean
            The currents assigned to each CC-CV cycle (without CV).
        2: C_charge
            The moved capacities during the charge segment(s). This is
            a list of the same length as charge, cv or discharge.
        3: U_charge
            The voltages during the charge segment(s). Length: same.
        4: C_discharge
            The moved capacities during the discharge segment(s).
            Length: same.
        5: U_discharge
            The voltages during the discharge segment(s). Length: same.
        6: C_evals
            Structurally the same as C_charge or C_discharge, this
            contains the moved capacities that were assigned to the mean
            voltages of charge and discharge cycle(s).
        7: U_means
            The mean voltages of each charge and discharge cycle.
    """

    I_mean = [0.0] * len(charge.timepoints)
    C_min = [0.0] * len(charge.timepoints)
    C_max = [0.0] * len(charge.timepoints)
    C_charge = []
    U_charge = []
    charge_splines = []
    corr = []
    C_discharge = []
    U_discharge = []
    discharge_splines = []
    OCV_fits = []
    C_evals = []
    U_means = []

    for i, (t, I, U) in enumerate(zip(cv.timepoints, cv.currents,
                                      cv.voltages)):
        Δt = np.array([t1 - t0 for (t0, t1) in zip(t[:-1], t[1:])])
        I_int = np.array([0.5 * (I0 + I1) for (I0, I1) in zip(I[:-1], I[1:])])
        C = [0.0] + list(np.cumsum(Δt * I_int) / 3600.0)
        corr.append(np.abs(C[-1]))

    for i, (t, I, U) in enumerate(zip(charge.timepoints, charge.currents,
                                      charge.voltages)):
        I_mean[i] += 0.5 * np.mean(I)
        Δt = np.array([t1 - t0 for (t0, t1) in zip(t[:-1], t[1:])])
        C = [0.0] + list(np.cumsum(Δt * np.array(I[1:])) / 3600.0)
        C = np.array(C) - np.min(C) + corr[i]
        C_charge.append(C)
        U_charge.append(U)
        C_min[i] = np.min(C)
        C_max[i] = np.max(C)
        charge_splines.append(smooth_fit(C, U,
                                         s=spline_smoothing))

    for i, (t, I, U) in enumerate(zip(discharge.timepoints, discharge.currents,
                                      discharge.voltages)):
        I_mean[i] -= 0.5 * np.mean(I)
        Δt = np.array([t1 - t0 for (t0, t1) in zip(t[:-1], t[1:])])
        C = [0.0] + list(np.cumsum(Δt * np.array(I[1:])) / 3600.0)
        C = np.array(C) - np.min(C) - corr[i]
        U = np.array(U)[C >= 0]
        C = C[C >= 0]
        C_discharge.append(C)
        U_discharge.append(U)
        C_min[i] = np.max([C_min[i], np.min(C)])
        C_max[i] = np.min([C_max[i], np.max(C)])
        discharge_splines.append(smooth_fit(C, U,
                                            s=spline_smoothing))

    for i, (min, max, c, d) in enumerate(zip(C_min, C_max, charge_splines,
                                             discharge_splines)):
        C_eval = np.linspace(min, max, eval_points)
        C_evals.append(C_eval)
        U_mean = 0.5 * (c(C_eval) + d(C_eval))
        U_means.append(U_mean)
        # diff = 0.5 * (c(C_eval) - d(C_eval))
        dummy_SOC = np.linspace(0.0, 1.0, eval_points)
        OCV_model = fit_OCV(dummy_SOC, U_mean, N=phases)
        OCV_fits.append(OCV_model)
        fit_SOC = np.linspace(*spline_SOC_range, eval_points)
        spline_OCV = smooth_fit(fit_SOC, inverse_OCV_fit_function(fit_SOC,
                                *OCV_model.fit, inverted=True), spline_order,
                                s=spline_smoothing)
        if parameters_print:
            print("Parameters of OCV fit function (" + str(I_mean[i]) + " A):")
            print("SOC range of data: " + repr(OCV_model.SOC_range))
            print("E₀: [" + ", ".join([str(x) for x in OCV_model.E_0]) + "]")
            print("a: [" + ", ".join([str(x) for x in OCV_model.a]) + "]")
            print("Δx: [" + ", ".join([str(x) for x in OCV_model.Δx]) + "]")
        if spline_order > 0:
            if parameters_print:
                print("Knots of interpolating spline:")
                print(spline_OCV.get_knots())
                print("Coefficients of this spline:")
                print(spline_OCV.get_coeffs())
            if spline_print is not None:
                print(verbose_spline_parameterization(
                    spline_OCV.get_coeffs(), spline_OCV.get_knots(),
                    spline_order, function_name=name, format=spline_print,
                    derivatives=1
                ))

    return (OCV_fits, I_mean, C_charge, U_charge, C_discharge, U_discharge,
            C_evals, U_means)


def calculate_desired_voltage(
        solution,
        t_eval,
        voltage_scale,
        overpotential,
        three_electrode=None,
        dimensionless_reference_electrode_location=0.5,
        parameters={},
):
    """!
    @param solution
        The pybamm.Solution object from which to calculate the voltage.
    @param t_eval
        The times at which to evaluate the "solution".
    @param voltage_scale
        The returned voltage gets divided by this value. For example,
        1e-3 would produce a plot in [mV].
    @param overpotential
        If True, only the overpotential of "solutions" gets plotted.
        Otherwise, the cell voltage (OCV + overpotential) is plotted.
    @param three_electrode
        With None, does nothing (i.e., cell potentials are used). If
        set to either 'positive' or 'negative', instead of cell
        potentials, the base for the displayed voltage will be the
        potential of the 'positive' or 'negative' electrode against a
        reference electrode. For placement of said reference electrode,
        please refer to "dimensionless_reference_electrode_location".
    @param dimensionless_reference_electrode_location
        The location of the reference electrode, given as a scalar
        between 0 (placed at the point where negative electrode and
        separator meet) and 1 (placed at the point where positive
        electrode and separator meet). Defaults to 0.5 (in the middle).
    @param parameters
        The parameter dictionary that was used for the simulation. Only
        needed for a three-electrode output.
    @return
        The array of the specified voltages over time.
    """

    if three_electrode:
        if parameters is None:
            raise ValueError(
                "To calculate the potentials for a three-electrode setup, "
                "the parameter dictionary of the simulation is needed. "
                "Give it as the keyword argument 'parameters'."
            )
        if three_electrode not in ["positive", "negative"]:
            raise ValueError(
                "'three_electrode' has to be either None, 'positive', or "
                "'negative'."
            )
        # When this issue gets completed, use it instead:
        # https://github.com/pybamm-team/PyBaMM/issues/2188
        L_n = parameters["Negative electrode thickness [m]"]
        L_s = parameters["Separator thickness [m]"]
        L_p = parameters["Positive electrode thickness [m]"]
        dimensional_location = (
            L_n + dimensionless_reference_electrode_location * L_s
        )
        x_working_electrode = (
            0 if three_electrode == "negative" else L_n + L_s + L_p
        )
        reference_electrode_potential = (
            solution["Electrolyte potential [V]"](
                t_eval, x=dimensional_location
            )
        )
        if overpotential:
            ocp = np.array([
                parameters[
                    three_electrode.capitalize() + " electrode OCP [V]"
                ](soc) for soc in solution[
                    "Average "
                    + three_electrode
                    + " particle concentration"
                ](t_eval)
            ])
            U = (
                solution[
                    three_electrode.capitalize()
                    + " electrode potential [V]"
                ](t_eval, x=x_working_electrode)
                - ocp
                - reference_electrode_potential
            ) / voltage_scale
        else:
            U = (
                solution[
                    three_electrode.capitalize()
                    + " electrode potential [V]"
                ](t_eval, x=x_working_electrode)
                - reference_electrode_potential
            ) / voltage_scale
    else:
        if overpotential:
            U = (
                solution["Voltage [V]"](t_eval)
                - solution["Bulk open-circuit voltage [V]"](t_eval, x=0),
            ) / voltage_scale
        else:
            U = solution["Voltage [V]"](t_eval) / voltage_scale
    return U


def solve_all_parameter_combinations(
    model,
    t_eval,
    parameters,
    parameters_to_try,
    submesh_types,
    var_pts,
    spatial_methods,
    full_factorial=True,
    **kwargs
):
    """!
    @param model
        The PyBaMM battery model that is to be solved.
    @param t_eval
        The timepoints in s at which this model is to be solved.
    @param parameters
        The model parameters as a dictionary.
    @param parameters_to_try
        A dictionary with the names of the model parameters as keys and
        lists of the values that are to be tried out for them as values.
    @param submesh_types
        The submeshes for discretization. See
        solversetup.spectral_mesh_pts_and_method.
    @param var_pts
        The number of discretization points. See
        solversetup.spectral_mesh_pts_and_method.
    @param spatial_methods
        The spatial methods for discretization. See
        solversetup.spectral_mesh_pts_and_method.
    @param full_factorial
        If True, all parameter combinations are tried out. If False,
        only each parameter is varied with the others staying fixed.
    @param kwargs
        The optional parameters for solversetup.solver_setup. See there.
    @return
        A 2-tuple with the model solution for parameters as first entry.
        The second entry mimics parameters_to_try with each entry in
        their lists replaced by the model solution for the corresponding
        parameter substitution. The second entry has one additional key
        "all parameters", where all parameters_to_try combinations are
        the value.
    """

    solutions = {}
    errorbars = {}
    free_parameters = list(parameters_to_try.keys())
    solver = solver_setup(
        model, parameters, submesh_types, var_pts, spatial_methods,
        free_parameters=free_parameters, **kwargs
    )
    input_parameters = {name: parameters[name] for name in free_parameters}
    solutions[model.name] = solver(t_eval, inputs=input_parameters)

    errorbars = {name: [
        solver(t_eval, inputs=p)
        for p in combine_parameters_to_try(input_parameters, {name: limits})[0]
    ] for name, limits in parameters_to_try.items()}

    errorbars["all parameters"] = [solutions[model.name]]
    for e in errorbars.values():
        errorbars["all parameters"].extend(e)
    if full_factorial:
        errorbars["all parameters"].extend([
            solver(t_eval, inputs=p)
            for p in combine_parameters_to_try(
                input_parameters, parameters_to_try
            )[0]
        ])

    return (solutions, errorbars)


def prepare_parameter_combinations(
    parameters,
    parameters_to_try,
    covariance,
    order_of_parameter_names,
    transform_parameters,
    confidence
):
    """!@brief Calculates all permutations of the parameter boundaries.
    @param parameters
        The model parameters as a dictionary.
    @param parameters_to_try
        A dictionary with the names of the model parameters as keys and
        lists of the values that are to be tried out for them as values.
        Mutually exclusive to "covariance".
    @param covariance
        A covariance matrix describing an estimation result of model
        parameters. Will be used to calculate parameters to try that
        together approximate the confidence ellipsoid. This confidence
        ellipsoid will be centered on "parameters".
        Mutually exclusive to "parameters_to_try".
    @param order_of_parameter_names
        A list of names from "parameters" that correspond to the order
        these parameters appear in the rows and columns of "covariance".
        Only needed when "covariance" is set.
    @param transform_parameters
        Optional transformations between the parameter space that is
        used for searching for optimal parameters and the model
        parameters. Any missing free parameter is not transformed.
        The values are 2-tuples. The first entry is a function taking
        the search space parameter and returning the model parameter.
        The second entry is the inverse function.
        For convenience, any value may also be one of the following:
         - 'none' => (identity, identity)
         - 'log' => (exp, log)
    @param confidence
        The confidence within the ellipsoid. Defaults to 0.95, i.e., the
        95% confidence ellipsoid.
    @return
        A 2-tuple with the individual parameter variations and then all
        permutations of them.
    """

    if not ((parameters_to_try is None) ^ (covariance is None)):
        raise ValueError(
            "Please set either 'parameters_to_try' or 'covariance'."
        )

    if parameters_to_try is not None:
        individual_bounds = {
            name: combine_parameters_to_try(
                parameters, {name: limits}
            )[0]
            for name, limits in parameters_to_try.items()
        }
        combinations = combine_parameters_to_try(
            parameters, parameters_to_try
        )[0]
    elif covariance is not None:
        individual_bounds = {
            "confidence semiaxes": approximate_confidence_ellipsoid(
                parameters,
                list(order_of_parameter_names),
                covariance,
                transform_parameters=transform_parameters,
                refinement=False,
                confidence=confidence
            )[0]
        }
        combinations = approximate_confidence_ellipsoid(
            parameters,
            list(order_of_parameter_names),
            covariance,
            transform_parameters=transform_parameters,
            refinement=True,
            confidence=confidence
        )[0][len(individual_bounds):]

    return individual_bounds, combinations


def parallel_simulator_with_setup(
    model,
    current_input,
    parameters,
    submesh_types,
    var_pts,
    spatial_methods,
    calc_esoh,
    inputs,
    t_eval,
    voltage_scale,
    overpotential,
    three_electrode,
    dimensionless_reference_electrode_location,
    kwargs
):
    par_var = copy.deepcopy(parameters)
    par_var.update(inputs)
    solver, callback = simulation_setup(
        model, current_input, par_var,
        submesh_types, var_pts, spatial_methods,
        **kwargs
    )
    solution = solver(
        check_model=False, calc_esoh=calc_esoh, callbacks=callback
    )
    variable = calculate_desired_voltage(
        solution,
        t_eval,
        voltage_scale,
        overpotential,
        three_electrode,
        dimensionless_reference_electrode_location,
        parameters
    )
    return variable


def simulate_all_parameter_combinations(
    model,
    current_input,
    submesh_types,
    var_pts,
    spatial_methods,
    parameters,
    parameters_to_try=None,
    covariance=None,
    order_of_parameter_names=None,
    additional_input_parameters=[],
    transform_parameters={},
    confidence=0.95,
    full_factorial=True,
    calc_esoh=False,
    voltage_scale=1.0,
    overpotential=False,
    three_electrode=None,
    dimensionless_reference_electrode_location=0.5,
    **kwargs
):
    """!
    @param model
        The PyBaMM battery model that is to be solved.
    @param current_input
        The list of battery operation conditions. See pybamm.Simulation.
    @param submesh_types
        The submeshes for discretization. See
        solversetup.spectral_mesh_pts_and_method.
    @param var_pts
        The number of discretization points. See
        solversetup.spectral_mesh_pts_and_method.
    @param spatial_methods
        The spatial methods for discretization. See
        solversetup.spectral_mesh_pts_and_method.
    @param parameters
        The model parameters as a dictionary.
    @param parameters_to_try
        A dictionary with the names of the model parameters as keys and
        lists of the values that are to be tried out for them as values.
        Mutually exclusive to "covariance".
    @param covariance
        A covariance matrix describing an estimation result of model
        parameters. Will be used to calculate parameters to try that
        together approximate the confidence ellipsoid. This confidence
        ellipsoid will be centered on "parameters".
        Mutually exclusive to "parameters_to_try".
    @param order_of_parameter_names
        A list of names from "parameters" that correspond to the order
        these parameters appear in the rows and columns of "covariance".
        Only needed when "covariance" is set.
    @param additional_input_parameters
        A list of the parameter names that are changed by any of the
        variable parameters, if "parameters" is a SubstitutionDict.
    @param transform_parameters
        Optional transformations between the parameter space that is
        used for searching for optimal parameters and the model
        parameters. Any missing free parameter is not transformed.
        The values are 2-tuples. The first entry is a function taking
        the search space parameter and returning the model parameter.
        The second entry is the inverse function.
        For convenience, any value may also be one of the following:
         - 'none' => (identity, identity)
         - 'log' => (exp, log)
    @param confidence
        The confidence within the ellipsoid. Defaults to 0.95, i.e., the
        95% confidence ellipsoid.
    @param full_factorial
        When "parameters_to_try" is set:
         - If True, all parameter combinations are tried out. If False,
           only each parameter is varied with the others staying fixed.
        When "covariance" is set:
         - If False, only the points on the semiaxes of the confidence
           ellipsoid constitute the parameters to try. If True, the
           centres of the faces of the polytope of these points get
           added to the parameters to try, projected onto the surface
           of the confidence ellipsoid.
    @param calc_esoh
        Passed on to pybamm.Simulator, see there.
    @param kwargs
        The optional parameters for solversetup.solver_setup. See there.
    @return
        A 2-tuple with the model solution for parameters as first entry.
        The second entry mimics parameters_to_try with each entry in
        their lists replaced by the model solution for the corresponding
        parameter substitution. The second entry has one additional key
        "all parameters", where all parameters_to_try combinations are
        the value.
    """

    individual_bounds, combinations = prepare_parameter_combinations(
        parameters,
        parameters_to_try,
        covariance,
        order_of_parameter_names,
        transform_parameters,
        confidence
    )

    if parameters_to_try is not None:
        if isinstance(parameters, SubstitutionDict):
            free_parameters = (
                list(parameters_to_try.keys())
                + list(additional_input_parameters)
            )
        else:
            free_parameters = list(parameters_to_try.keys())
    elif covariance is not None:
        if isinstance(parameters, SubstitutionDict):
            free_parameters = (
                list(order_of_parameter_names)
                + list(additional_input_parameters)
            )
        else:
            free_parameters = list(order_of_parameter_names)

    individual_bounds = {
        name: [
            {
                free_name: parameters_variation[free_name]
                for free_name in free_parameters
            }
            for parameters_variation in bounds
        ]
        for name, bounds in individual_bounds.items()
    }
    combinations = [
        {
            free_name: parameters_variation[free_name]
            for free_name in free_parameters
        }
        for parameters_variation in combinations
    ]

    solutions = {}
    errorbars = {}

    solver, callback = simulation_setup(
        model, current_input, parameters,
        submesh_types, var_pts, spatial_methods,
        **kwargs
    )
    solutions[model.name] = solver(
        check_model=False, calc_esoh=calc_esoh, callbacks=callback
    )
    t_eval = solutions[model.name].t

    parallel_arguments = []
    indices_of_individual_bounds = {}
    index_counter = 0
    for name in individual_bounds:
        new_parallel_arguments = [
            (
                copy.deepcopy(model),
                current_input,
                copy.deepcopy(parameters),
                submesh_types,
                var_pts,
                spatial_methods,
                calc_esoh,
                ib,
                t_eval,
                voltage_scale,
                overpotential,
                three_electrode,
                dimensionless_reference_electrode_location,
                kwargs,
            )
            for ib in individual_bounds[name]
        ]
        indices_of_individual_bounds[name] = [i for i in range(
            index_counter, index_counter + len(new_parallel_arguments)
        )]
        index_counter += len(new_parallel_arguments)
        parallel_arguments.extend(new_parallel_arguments)
    with Pool() as p:
        all_errorbars = p.starmap(
            parallel_simulator_with_setup, parallel_arguments
        )
    for name in individual_bounds:
        errorbars[name] = [
            all_errorbars[i] for i in indices_of_individual_bounds[name]
        ]

    errorbars["all parameters"] = [solutions[model.name]]
    for e in errorbars.values():
        errorbars["all parameters"].extend(e)
    if full_factorial:
        parallel_arguments = [
            (
                copy.deepcopy(model),
                current_input,
                copy.deepcopy(parameters),
                submesh_types,
                var_pts,
                spatial_methods,
                calc_esoh,
                c,
                t_eval,
                voltage_scale,
                overpotential,
                three_electrode,
                dimensionless_reference_electrode_location,
                kwargs,
            )
            for c in combinations
        ]
        with Pool() as p:
            errorbars["all parameters"].extend(p.starmap(
                parallel_simulator_with_setup, parallel_arguments
            ))

    return (solutions, errorbars)



"""!@package ep_bolfi.models.electrolyte
Contains a PyBaMM-compatible electrolyte model.
"""

import pybamm
import ep_bolfi.models.standard_parameters as standard_parameters
from ep_bolfi.models.standard_parameters import (
    OCVₙ_ref, OCVₚ_ref, OCVₙ, OCVₚ,
    cₙ, cₚ,
    iₛₑₙ_0, iₛₑₚ_0,
    zₙ, zₚ,
    αₙₙ, αₚₙ, αₙₚ, αₚₚ,
    σₙ, σₚ,
    Rₙ, Rₚ,
    cₑ_typ, cₑ_init,
    εₛ, εₙ_scalar, εₛ_scalar, εₚ_scalar,
    βₑₛ_scalar,
    Dₑ, κₑ, κₑ_hat, t_plus, one_plus_dlnf_dlnc,
    Lₙ_dim, Lₚ_dim, L_dim, Lₙ, Lₚ,
    Cₑ, Cᵣₙ, Cᵣₚ,
    γₑ, γₙ, γₚ,
    C, A, F, τᵈ, capacity,
    T_init, thermal_voltage,
    I_extern,
    Uₗ, Uᵤ,
)
# Reset the PyBaMM colour scheme.
import matplotlib.pyplot as plt
plt.style.use("default")


class Electrolyte_internal(pybamm.BaseSubModel):
    """!@brief Defining equations for a symmetric Li cell with electrolyte.
    Reference
    ----------
    SG Marquis, V Sulzer, R Timms, CP Please and SJ Chapman.
    “An asymptotic derivation of a single particle model with
    electrolyte”. Journal of The Electrochemical Society,
    166(15):A3693–A3706, 2019
    """

    def __init__(
        self,
        param,
        pybamm_control=False,
        options={},
        build=True
    ):
        """!@brief Sets the model properties.
        @param param
            A class containing all the relevant parameters for this
            model. For example, models.standard_parameters represents a
            valid choice for this parameter.
        @param pybamm_control
            Per default False, which indicates that the current is given
            as a function. If set to True, this model is compatible with
            PyBaMM experiments, e.g. CC-CV simulations. The current is
            then a variable and it or voltage can be fixed functions.
        @param options
            Not used; only here for compatibility with the base class.
        @param build
            Not used; only here for compatibility with the base class.
        """

        super().__init__(param)
        ## Current is fixed if False and a variable if True.
        self.pybamm_control = pybamm_control

    def get_fundamental_variables(self):
        """!@brief Builds all relevant model variables' symbols.
        @return
            A dictionary with the variables' names as keys and their
            symbols (of type pybamm.Symbol) as values.
        """

        # Define the variables.
        t = pybamm.Variable("Time [h]")
        if not self.pybamm_control:
            # I_cell = I_extern
            Q = pybamm.Variable("Discharge capacity [A.h]")

        # Variables that vary spatially are created with a domain
        cₑₛ = pybamm.Variable(
            "Separator electrolyte concentration", domain="separator"
        )
        cₑ = cₑₛ

        # Electrolyte potential
        φₑₛ = pybamm.Variable(
            "Separator electrolyte potential", domain="separator"
        )
        φₑ = φₑₛ

        SOCₙ = pybamm.Scalar(0)
        SOCₚ = pybamm.Scalar(0)
        # total_SOCₙ = pybamm.Scalar(0)
        # total_SOCₚ = pybamm.Scalar(0)
        SOCₙ_surf = pybamm.Scalar(0)
        SOCₚ_surf = pybamm.Scalar(0)

        # Constant temperature
        T = T_init
        # Porosity
        # ε = εₛ
        # Tortuosity
        εᵝ = εₛ**βₑₛ_scalar

        # Interfacial reactions

        ηₙ = pybamm.Variable("Negative electrode overpotential")
        ηₚ = pybamm.Variable("Positive electrode overpotential")

        # Butler-Volmer reaction fluxes.
        # Note: not using the simplified version with sinh, which is
        # fixed to symmetric charge-transfer, is significantly slower.
        # Note: γₙ / Cᵣₙ does not depend on cₙ; it cancels out.
        iₛₑₙ = (
            (γₙ / Cᵣₙ)
            * iₛₑₙ_0(pybamm.boundary_value(cₑₛ, "left"), 0.5, T)
            * (pybamm.exp(αₙₙ * zₙ * ηₙ) - pybamm.exp(-αₚₙ * zₙ * ηₙ))
        )
        iₛₑₛ = pybamm.PrimaryBroadcast(0, "separator")
        iₛₑₚ = (
            (γₚ / Cᵣₚ) * iₛₑₚ_0(pybamm.boundary_value(cₑₛ, "right"), 0.5, T)
            * (pybamm.exp(αₙₚ * zₚ * ηₚ) - pybamm.exp(-αₚₚ * zₚ * ηₚ))
        )
        iₛₑ = iₛₑₛ

        # Current in the electrolyte
        iₑ = εᵝ * κₑ_hat * κₑ(cₑ, T) * (
            2 * one_plus_dlnf_dlnc(cₑ)
            * (1 - t_plus(cₑ)) * pybamm.grad(cₑ) / cₑ
            - pybamm.grad(φₑ)
        )

        # Calculations for overpotential
        total_SOCₙ = pybamm.Scalar(0)
        total_SOCₚ = pybamm.Scalar(0)
        # The `variables` dictionary contains all variables that might
        # be useful for visualising the solution of the model
        variables = {
            "Negative electrode SOC": total_SOCₙ,
            "Negative particle concentration": SOCₙ,
            "Negative particle concentration [mol.m-3]": SOCₙ * cₙ,
            "Negative particle surface concentration": SOCₙ_surf,
            "Negative particle surface concentration [mol.m-3]":
                SOCₙ_surf * cₙ,
            "Negative electrolyte concentration": pybamm.Scalar(0),
            "Negative electrolyte concentration [mol.m-3]": pybamm.Scalar(0),
            "Separator electrolyte concentration": cₑₛ,
            "Separator electrolyte concentration [mol.m-3]": cₑₛ * cₑ_typ,
            "Positive electrolyte concentration": pybamm.Scalar(0),
            "Positive electrolyte concentration [mol.m-3]": pybamm.Scalar(0),
            "Electrolyte concentration": cₑ,
            "Electrolyte concentration [mol.m-3]": cₑ * cₑ_typ,
            "Positive electrode SOC": total_SOCₚ,
            "Positive particle concentration": SOCₚ,
            "Positive particle concentration [mol.m-3]": SOCₚ * cₚ,
            "Positive particle surface concentration": SOCₚ_surf,
            "Positive particle surface concentration [mol.m-3]":
                SOCₚ_surf * cₚ,
            "Negative electrode interface current density": iₛₑₙ,
            "Positive electrode interface current density": iₛₑₚ,
            "Interface current density": iₛₑ,
            "Interface current density [A.m-2]": iₛₑₛ,
            "Negative electrode overpotential": ηₙ,
            "Negative electrode overpotential [V]": ηₙ * thermal_voltage,
            "Electrolyte current": iₑ,
            # For computation, use the version without NaN values.
            # "Negative electrolyte potential": φₑₙ,
            # "Negative electrolyte potential [V]": φₑₙ * thermal_voltage,
            "Separator electrolyte potential": φₑₛ,
            "Separator electrolyte potential [V]": φₑₛ * thermal_voltage,
            # "Positive electrolyte potential": φₑₚ,
            # "Positive electrolyte potential [V]": φₑₚ * thermal_voltage,
            "Electrolyte potential": φₑ,
            "Electrolyte potential [V]": φₑ * thermal_voltage,
            "Positive electrode overpotential": ηₚ,
            "Positive electrode overpotential [V]": ηₚ * thermal_voltage,
            # The following variable names are for PyBaMM compatibility.
            # In particular, they are required if pybamm_control==True.
            "Time [h]": t,
            "C-rate": C,
            "Negative electrode capacity [A.h]":
                (1 - εₙ_scalar) * Lₙ_dim * cₙ * zₙ * F * A,
            "Positive electrode capacity [A.h]":
                (1 - εₚ_scalar) * Lₚ_dim * cₚ * zₚ * F * A,
            "Negative electrode open circuit potential": OCVₙ(SOCₙ_surf, T),
            "Negative electrode open circuit potential [V]":
                OCVₙ_ref + thermal_voltage * OCVₙ(SOCₙ_surf, T),
            "Negative electrode open circuit potential over time [V]":
                OCVₙ_ref + thermal_voltage * OCVₙ(total_SOCₙ, T),
            "Positive electrode open circuit potential": OCVₚ(SOCₚ_surf, T),
            "Positive electrode open circuit potential [V]":
                OCVₚ_ref + thermal_voltage * OCVₚ(SOCₚ_surf, T),
            "Positive electrode open circuit potential over time [V]":
                OCVₚ_ref + thermal_voltage * OCVₚ(total_SOCₚ, T),
            "X-averaged negative electrode open circuit potential":
                OCVₙ(SOCₙ_surf, T),
            "X-averaged negative electrode open circuit potential [V]":
                OCVₙ_ref + thermal_voltage * OCVₙ(SOCₙ_surf, T),
            "X-averaged positive electrode open circuit potential":
                OCVₚ(SOCₚ_surf, T),
            "X-averaged positive electrode open circuit potential [V]":
                OCVₚ_ref + thermal_voltage * OCVₚ(SOCₚ_surf, T),
            "X-averaged negative electrode reaction overpotential":
                ηₙ,
            "X-averaged negative electrode reaction overpotential [V]":
                thermal_voltage * ηₙ,
            "X-averaged positive electrode reaction overpotential":
                ηₚ,
            "X-averaged positive electrode reaction overpotential [V]":
                thermal_voltage * ηₚ,
            "X-averaged negative electrode sei film overpotential":
                pybamm.Scalar(0),
            "X-averaged negative electrode sei film overpotential [V]":
                pybamm.Scalar(0),
            "X-averaged positive electrode sei film overpotential":
                pybamm.Scalar(0),
            "X-averaged positive electrode sei film overpotential [V]":
                pybamm.Scalar(0),
            "Nominal cell capacity [A.h]": capacity,
            "Loss of active material in negative electrode [%]":
                pybamm.Scalar(0),
            "Loss of active material in positive electrode [%]":
                pybamm.Scalar(0),
            "Loss of lithium inventory [%]": pybamm.Scalar(0),
            "Loss of lithium inventory, including electrolyte [%]":
                pybamm.Scalar(0),
            "Total lithium lost [mol]": pybamm.Scalar(0),
            "Total lithium lost from particles [mol]": pybamm.Scalar(0),
            "Total lithium lost from electrolyte [mol]": pybamm.Scalar(0),
            "Loss of lithium to negative electrode SEI [mol]":
                pybamm.Scalar(0),
            "Loss of lithium to positive electrode SEI [mol]":
                pybamm.Scalar(0),
            "Loss of lithium to negative electrode lithium plating [mol]":
                pybamm.Scalar(0),
            "Loss of lithium to positive electrode lithium plating [mol]":
                pybamm.Scalar(0),
            "Loss of capacity to negative electrode SEI [A.h]":
                pybamm.Scalar(0),
            "Loss of capacity to positive electrode SEI [A.h]":
                pybamm.Scalar(0),
            "Loss of capacity to negative electrode lithium plating [A.h]":
                pybamm.Scalar(0),
            "Loss of capacity to positive electrode lithium plating [A.h]":
                pybamm.Scalar(0),
            "Total lithium lost to side reactions [mol]": pybamm.Scalar(0),
            "Total capacity lost to side reactions [A.h]": pybamm.Scalar(0),
            "Loss of lithium to SEI [mol]": pybamm.Scalar(0),
            "Loss of lithium to lithium plating [mol]": pybamm.Scalar(0),
            "Loss of capacity to SEI [A.h]": pybamm.Scalar(0),
            "Loss of capacity to lithium plating [A.h]": pybamm.Scalar(0),
            # These parameters are only a placeholder.
            "Total lithium [mol]": pybamm.Scalar(0),
            "Total lithium in electrolyte [mol]": pybamm.Scalar(0),
            "Total lithium in negative electrode [mol]": pybamm.Scalar(0),
            "Total lithium in positive electrode [mol]": pybamm.Scalar(0),
            "Total lithium in particles [mol]": pybamm.Scalar(0),
            "Local ECM resistance [Ohm]": pybamm.Scalar(0),
        }
        if not self.pybamm_control:
            variables["Discharge capacity [A.h]"] = Q
        return variables

    def get_coupled_variables(self, variables):
        """!@brief Builds all model symbols that rely on other models.
        @param variables
            A dictionary containing at least all variable symbols that
            are required for the variable symbols built here.
        @return
            A dictionary with the new variables' names as keys and their
            symbols (of type pybamm.Symbol) as values.
        """

        if self.pybamm_control:
            I_cell = variables["Total current density"]
        else:
            I_cell = I_extern
            # Q = pybamm.Variable("Discharge capacity [A.h]")

        # Electrode potential
        # Note: the "porosity" of the electrode is 1.
        # φₛₙ = -I_cell * (Lₙ / σₙ) * xₙ / Lₙ
        φₛₙ = -I_cell * (Lₙ / σₙ)
        φₛₚ_contribution = I_cell * (Lₚ / σₚ)

        φₑₛ = variables["Separator electrolyte potential"]
        cₑₛ = variables["Separator electrolyte concentration"]

        # Current in the solid
        iₛₙ = I_cell
        iₛₚ = I_cell

        ηₙ = variables["Negative electrode overpotential"]
        ηₚ = variables["Positive electrode overpotential"]

        φₛₚ = (
            φₛₙ
            - ηₙ
            - pybamm.boundary_value(φₑₛ, "left")
            - pybamm.log(pybamm.boundary_value(cₑₛ, "left")) / zₙ
            + pybamm.boundary_value(φₑₛ, "right")
            + pybamm.log(pybamm.boundary_value(cₑₛ, "right")) / zₚ
            + ηₚ
        )
        # Cell voltage
        voltage = φₛₚ - φₛₚ_contribution
        voltage_dim = voltage * thermal_voltage
        overpotential_dim = voltage_dim
        overpotential = overpotential_dim / thermal_voltage

        return {
            "Negative electrode potential": φₛₙ,
            "Negative electrode potential [V]": φₛₙ * thermal_voltage,
            "Positive electrode potential": φₛₚ,
            "Positive electrode potential [V]":
                OCVₚ_ref - OCVₙ_ref + φₛₚ * thermal_voltage,
            "Negative electrode current": iₛₙ,
            "Positive electrode current": iₛₚ,
            # The following variable names are for PyBaMM compatibility.
            # In particular, they are required if pybamm_control==True.
            "Terminal voltage": voltage,
            "Terminal voltage [V]": voltage_dim,
            "Battery voltage [V]": voltage_dim,
            "Total overpotential": overpotential,
            "Total overpotential [V]": overpotential_dim,
            "X-averaged negative electrode ohmic losses":
                φₛₙ,
            "X-averaged negative electrode ohmic losses [V]":
                thermal_voltage * φₛₙ,
            "X-averaged positive electrode ohmic losses":
                φₛₚ,
            "X-averaged positive electrode ohmic losses [V]":
                thermal_voltage * φₛₚ,
        }

    def set_rhs(self, variables):
        """!@brief Sets up the right-hand-side equations in self.rhs. """

        t = variables["Time [h]"]
        if self.pybamm_control:
            I_cell = variables["Total current density"]
        else:
            I_cell = I_extern
            Q = variables["Discharge capacity [A.h]"]
        iₑ = variables["Electrolyte current"]
        iₛₑ = variables["Interface current density"]
        cₑ = variables["Electrolyte concentration"]
        # Constant temperature
        T = T_init
        # Porosity
        ε = εₛ
        # Tortuosity
        εᵝ = εₛ**βₑₛ_scalar

        self.rhs[t] = τᵈ / 3600
        if not self.pybamm_control:
            self.rhs[Q] = C * I_cell * τᵈ / 3600
        # Electrolyte concentration
        Nₑ = -εᵝ * Dₑ(cₑ, T) * pybamm.grad(cₑ) + (Cₑ / γₑ) * t_plus(cₑ) * iₑ
        self.rhs[cₑ] = (1 / ε) * (-pybamm.div(Nₑ) / Cₑ + iₛₑ / γₑ)

    def set_algebraic(self, variables):
        """!@brief Sets up the algebraic equations in self.algebraic. """

        if self.pybamm_control:
            I_cell = variables["Total current density"]
        else:
            I_cell = I_extern
        ηₙ = variables["Negative electrode overpotential"]
        ηₚ = variables["Positive electrode overpotential"]
        iₛₑₙ = variables["Negative electrode interface current density"]
        iₛₑₚ = variables["Positive electrode interface current density"]
        iₛₑ = variables["Interface current density"]
        iₑ = variables["Electrolyte current"]
        φₑ = variables["Electrolyte potential"]

        self.algebraic[ηₙ] = iₛₑₙ - I_cell
        self.algebraic[ηₚ] = iₛₑₚ + I_cell
        self.algebraic[φₑ] = pybamm.div(iₑ) - iₛₑ

    def set_boundary_conditions(self, variables):
        """!@brief Sets the (self.)boundary(_)conditions. """

        if self.pybamm_control:
            I_cell = variables["Total current density"]
        else:
            I_cell = I_extern
        ηₙ = variables["Negative electrode overpotential"]
        φₛₙ = variables["Negative electrode potential"]
        φₑ = variables["Electrolyte potential"]
        cₑ = variables["Electrolyte concentration"]
        # Constant temperature
        T = T_init

        cₑₙ = pybamm.boundary_value(cₑ, "left")
        cₑₚ = pybamm.boundary_value(cₑ, "right")

        # At one electrode surface, a Dirichlet boundary condition is
        # required. At the other electrode surface, a Neumann boundary
        # condition is required to determine the slope.
        self.boundary_conditions[φₑ] = {
            "left": (φₛₙ - ηₙ - pybamm.log(cₑₙ) / zₙ, "Dirichlet"),
            "right": (
                (
                    2 * one_plus_dlnf_dlnc(cₑₚ) * (1 - t_plus(cₑₚ)) * (
                        -Cₑ * (1 - t_plus(cₑₚ)) * I_cell / (
                            γₑ * εₛ_scalar**βₑₛ_scalar * Dₑ(cₑₚ, T)
                        )
                    ) / cₑₚ
                ) - I_cell / (εₛ_scalar**βₑₛ_scalar * κₑ_hat * κₑ(cₑₚ, T)),
                "Neumann"
            )
        }
        self.boundary_conditions[cₑ] = {
            "left": (
                -Cₑ * (1 - t_plus(cₑₙ)) * I_cell
                / (γₑ * εₛ_scalar**βₑₛ_scalar * Dₑ(cₑₙ, T)),
                "Neumann"
            ),
            "right": (
                -Cₑ * (1 - t_plus(cₑₚ)) * I_cell
                / (γₑ * εₛ_scalar**βₑₛ_scalar * Dₑ(cₑₚ, T)),
                "Neumann"
            ),
        }

    def set_initial_conditions(self, variables):
        """!@brief Sets the (self.)initial(_)conditions. """

        # Initial conditions for rhs: actual start values.
        # Initial conditions for algebraic: help for root-finding.
        t = variables["Time [h]"]
        if not self.pybamm_control:
            Q = variables["Discharge capacity [A.h]"]
        ηₙ = variables["Negative electrode overpotential"]
        ηₚ = variables["Positive electrode overpotential"]
        φₑ = variables["Electrolyte potential"]
        cₑ = variables["Electrolyte concentration"]

        self.initial_conditions[t] = pybamm.Scalar(0)
        if not self.pybamm_control:
            self.initial_conditions[Q] = pybamm.Scalar(0)
        self.initial_conditions[ηₙ] = pybamm.Scalar(0)
        self.initial_conditions[ηₚ] = pybamm.Scalar(0)
        self.initial_conditions[φₑ] = pybamm.Scalar(0)
        self.initial_conditions[cₑ] = cₑ_init

    def set_events(self, variables):
        """!@brief Sets up the termination switches in self.events. """

        SOCₙ_surf = pybamm.Scalar(0)
        SOCₚ_surf = pybamm.Scalar(0)
        cₑ = variables["Electrolyte concentration"]
        voltage_dim = variables["Terminal voltage [V]"]
        # Events specify points at which a solution should terminate
        self.events += [
            pybamm.Event(
                "Minimum negative particle surface concentration",
                pybamm.min(SOCₙ_surf) - 0.001,
            ),
            pybamm.Event(
                "Maximum negative particle surface concentration",
                (1 - 0.001) - pybamm.max(SOCₙ_surf),
            ),
            pybamm.Event(
                "Minimum positive particle surface concentration",
                pybamm.min(SOCₚ_surf) - 0.001,
            ),
            pybamm.Event(
                "Maximum positive particle surface concentration",
                (1 - 0.001) - pybamm.max(SOCₚ_surf),
            ),
            pybamm.Event(
                "Zero electrolyte concentration cut-off",
                pybamm.min(cₑ) - 0.002
            ),
            pybamm.Event("Minimum voltage", voltage_dim - Uₗ),
            pybamm.Event("Maximum voltage", voltage_dim - Uᵤ),
        ]


class Electrolyte(pybamm.BaseBatteryModel):
    """!@brief Electrolyte model assuming a symmetric Li-metal cell. """

    def __init__(
        self,
        pybamm_control=False,
        name="electrolyte",
        options={},
        build=True
    ):
        """!@brief Sets up an electrolyte model usable by PyBaMM.
        @param pybamm_control
            Per default False, which indicates that the current is given
            as a function. If set to True, this model is compatible with
            PyBaMM experiments, e.g. CC-CV simulations. The current is
            then a variable and it or voltage can be fixed functions.
        @param name
            The optional name of the model. Default is "electrolyte".
        @param options
            Used for external circuit if pybamm_control is True.
        @param build
            Per default True, which builds the model equations right
            away. If set to False, they remain as symbols for later.
        """

        super().__init__(name=name)
        ## Current is fixed if False and a variable if True.
        self.pybamm_control = pybamm_control
        pybamm.citations.register("Marquis2019")
        self.options = options

        ## Contains all the relevant parameters for this model.
        self.param = standard_parameters
        ## Non-dimensionalization timescale.
        self.timescale = τᵈ
        ## Non-dimensionalization length scales.
        self.length_scales = {
            "negative electrode": L_dim,
            "separator": L_dim,
            "positive electrode": L_dim,
            "negative particle": Rₙ,
            "positive particle": Rₚ,
            # "current collector y": self.param.L_y,
            # "current collector z": self.param.L_z,
        }
        self.set_standard_output_variables()

        if self.pybamm_control:
            self.set_external_circuit_submodel()
        self.submodels["internal"] = Electrolyte_internal(
            self.param, self.pybamm_control
        )

        if build:
            self.build_model()

    def set_standard_output_variables(self):
        """!@brief Adds "the horizontal axis" to self.variables.
        Don't use the super() version of this function, as it
        introduces keys with integer values in self.variables.
        """

        # super().set_standard_output_variables()
        var = pybamm.standard_spatial_vars
        param = self.param
        self.variables.update(
            {
                "Time": pybamm.t,
                "Time [s]": pybamm.t * self.timescale,
                "Time [min]": pybamm.t * self.timescale / 60,
                "Time [h]": pybamm.t * self.timescale / 3600,
                # "y": var.y,
                # "y [m]": var.y * param.L_y,
                # "z": var.z,
                # "z [m]": var.z * param.L_z,
                "x_s": var.x_s,
                "x_s [m]": var.x_s * param.L_dim,
            }
        )
        x = var.x_s
        self.variables.update(
            {
                "x": x,
                "x [m]": x * param.L_dim,
            }
        )

    def new_copy(self, build=True):
        """!@brief Create an empty copy with identical options.
        @param build
            If True, the new model gets built right away. This is the
            default behavior. If set to False, it remains as symbols.
        @return
            The copy of this model.
        """

        new_model = self.__class__(self.pybamm_control,
                                   self.name, build=False)
        # update submodels
        new_model.submodels = self.submodels
        # clear submodel equations to avoid weird conflicts
        for submodel in self.submodels.values():
            submodel._rhs = {}
            submodel._algebraic = {}
            submodel._initial_conditions = {}
            submodel._boundary_conditions = {}
            submodel._variables = {}
            submodel._events = []

        # now build
        if build:
            new_model.build_model()
        new_model.use_jacobian = self.use_jacobian
        # new_model.use_simplify = self.use_simplify
        new_model.convert_to_format = self.convert_to_format
        new_model.timescale = self.timescale
        new_model.length_scales = self.length_scales
        """
        replacer = pybamm.SymbolReplacer({})
        return replacer.process_model(self, inplace=False)
        """
        return new_model

    def set_voltage_variables(self):
        """!@brief Adds voltage-specific variables to self.variables.
        Override this inherited function, since it adds superfluous
        variables.
        """

        pass

    @property
    def default_geometry(self):
        """!@brief Override: corrects the geometry for half-cells. """

        geometry = super().default_geometry
        del geometry["negative electrode"]
        del geometry["negative particle"]
        del geometry["positive electrode"]
        del geometry["positive particle"]
        return geometry
